#!/usr/bin/env python3
"""Manages CoLRev search sources."""
from __future__ import annotations

import typing
from dataclasses import asdict

import colrev.operation

# pylint: disable=too-few-public-methods


class SearchSources:
    """SearchSources (including academic databases, citation searches, PDF files)"""

    def __init__(self, *, review_manager: colrev.review_manager.ReviewManager) -> None:
        package_manager = review_manager.get_package_manager()
        check_operation = colrev.operation.CheckOperation(review_manager=review_manager)

        self.all_available_packages_names = package_manager.discover_packages(
            package_type=colrev.env.package_manager.PackageEndpointType.search_source,
            installed_only=True,
        )
        # Note: class-objects only (instantiate_objects) for heuristics
        self.all_available_packages = package_manager.load_packages(
            package_type=colrev.env.package_manager.PackageEndpointType.search_source,
            selected_packages=[
                {
                    "endpoint": k,
                }
                for k in list(self.all_available_packages_names.keys())
            ],
            operation=check_operation,
            instantiate_objects=False,
        )

        self.packages: dict[str, typing.Any] = package_manager.load_packages(
            package_type=colrev.env.package_manager.PackageEndpointType.search_source,
            selected_packages=[asdict(s) for s in review_manager.settings.sources],
            operation=check_operation,
            ignore_not_available=False,
        )



#! /usr/bin/env python
"""Template for a custom PDFGet PackageEndpoint"""
from __future__ import annotations

from typing import TYPE_CHECKING

import zope.interface
from dacite import from_dict

import colrev.operation

if TYPE_CHECKING:
    import colrev.ops.pdf_get


# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.env.package_manager.PDFGetPackageEndpointInterface)
class CustomPDFGet:
    """Class for custom pdf-get scripts"""

    def __init__(
        self,
        *,
        pdf_get_operation: colrev.ops.pdf_get.PDFGet,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = from_dict(
            data_class=colrev.env.package_manager.DefaultSettings, data=settings
        )

    def get_pdf(
        self,
        pdf_get_operation: colrev.ops.pdf_get.PDFGet,  # pylint: disable=unused-argument
        record: colrev.record.Record,
    ) -> colrev.record.Record:
        """Get the PDF"""

        record.data["file"] = "filepath"
        pdf_get_operation.import_pdf(record=record)

        return record



#! /usr/bin/env python
"""SearchSource: Scopus"""
from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.load_utils_bib
import colrev.ops.search
import colrev.record

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class ScopusSearchSource(JsonSchemaMixin):
    """SearchSource for scopus"""

    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "url"
    search_type = colrev.settings.SearchType.DB
    api_search_supported = False
    ci_supported: bool = False
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.supported
    short_name = "Scopus"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/ops/built_in/search_sources/scopus.md"
    )

    def __init__(
        self, *, source_operation: colrev.operation.Operation, settings: dict
    ) -> None:
        self.search_source = from_dict(data_class=self.settings_class, data=settings)
        self.quality_model = source_operation.review_manager.get_qm()

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for Scopus"""

        result = {"confidence": 0.0}
        if "source={Scopus}," in data:
            result["confidence"] = 1.0
            return result

        if "www.scopus.com" in data:
            if data.count("www.scopus.com") >= data.count("\n@"):
                result["confidence"] = 1.0

        return result

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.search.Search, params: str) -> None:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""
        raise NotImplementedError

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of Scopus"""

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Not implemented"""
        return record

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.ops.load_utils_bib.load_bib_file(
                load_operation=load_operation, source=self.search_source
            )
            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for Scopus"""

        if "conference" == record.data["ENTRYTYPE"]:
            record.data["ENTRYTYPE"] = "inproceedings"

        if "book" == record.data["ENTRYTYPE"]:
            if "journal" in record.data and "booktitle" not in record.data:
                record.rename_field(key="title", new_key="booktitle")
                record.rename_field(key="journal", new_key="title")

        if "colrev.scopus.document_type" in record.data:
            if record.data["colrev.scopus.document_type"] == "Conference Paper":
                record.change_entrytype(
                    new_entrytype="inproceedings", qm=self.quality_model
                )

            elif record.data["colrev.scopus.document_type"] == "Conference Review":
                record.change_entrytype(
                    new_entrytype="proceedings", qm=self.quality_model
                )

            elif record.data["colrev.scopus.document_type"] == "Article":
                record.change_entrytype(new_entrytype="article", qm=self.quality_model)

            record.remove_field(key="colrev.scopus.document_type")

        if (
            "colrev.scopus.Start_Page" in record.data
            and "colrev.scopus.End_Page" in record.data
        ):
            if (
                record.data["colrev.scopus.Start_Page"] != "nan"
                and record.data["colrev.scopus.End_Page"] != "nan"
            ):
                record.data["pages"] = (
                    record.data["colrev.scopus.Start_Page"]
                    + "--"
                    + record.data["colrev.scopus.End_Page"]
                )
                record.data["pages"] = record.data["pages"].replace(".0", "")
                record.remove_field(key="colrev.scopus.Start_Page")
                record.remove_field(key="colrev.scopus.End_Page")

        if "colrev.scopus.note" in record.data:
            if "cited By " in record.data["colrev.scopus.note"]:
                record.rename_field(key="colrev.scopus.note", new_key="cited_by")
                record.data["cited_by"] = record.data["cited_by"].replace(
                    "cited By ", ""
                )

        if "author" in record.data:
            record.data["author"] = record.data["author"].replace("; ", " and ")

        record.remove_field(key="colrev.scopus.source")

        return record



#! /usr/bin/env python
"""SearchSource: Europe PMC"""
from __future__ import annotations

import json
import typing
from copy import deepcopy
from dataclasses import dataclass
from multiprocessing import Lock
from pathlib import Path
from sqlite3 import OperationalError
from typing import Optional
from urllib.parse import quote
from urllib.parse import urlparse
from xml.etree.ElementTree import Element  # nosec

import defusedxml
import requests
import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin
from defusedxml.ElementTree import fromstring
from thefuzz import fuzz

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.load_utils_bib
import colrev.ops.search
import colrev.record
import colrev.settings

# defuse std xml lib
defusedxml.defuse_stdlib()


# pylint: disable=duplicate-code
# pylint: disable=unused-argument


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class EuropePMCSearchSource(JsonSchemaMixin):
    """SearchSource for Europe PMC"""

    # settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "europe_pmc_id"
    search_type = colrev.settings.SearchType.DB
    endpoint = "colrev.europe_pmc"
    api_search_supported = True
    ci_supported: bool = True
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.supported
    short_name = "Europe PMC"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/ops/built_in/search_sources/europe_pmc.md"
    )
    __europe_pmc_md_filename = Path("data/search/md_europe_pmc.bib")
    __SOURCE_URL = "https://www.ebi.ac.uk/europepmc/webservices/rest/article/"
    __next_page_url: str = ""

    @dataclass
    class EuropePMCSearchSourceSettings(colrev.settings.SearchSource, JsonSchemaMixin):
        """Settings for EuropePMCSearchSource"""

        # pylint: disable=too-many-instance-attributes
        endpoint: str
        filename: Path
        search_type: colrev.settings.SearchType
        search_parameters: dict
        comment: typing.Optional[str]

        _details = {
            "search_parameters": {
                "tooltip": "Currently supports a scope item "
                "with venue_key and journal_abbreviated fields."
            },
        }

    settings_class = EuropePMCSearchSourceSettings

    def __init__(
        self,
        *,
        source_operation: colrev.operation.Operation,
        settings: Optional[dict] = None,
    ) -> None:
        self.review_manager = source_operation.review_manager
        if settings:
            # EuropePMC as a search_source
            self.search_source = from_dict(
                data_class=self.settings_class, data=settings
            )
        else:
            # EuropePMC as an md-prep source
            europe_pmc_md_source_l = [
                s
                for s in source_operation.review_manager.settings.sources
                if s.filename == self.__europe_pmc_md_filename
            ]
            if europe_pmc_md_source_l:
                self.search_source = europe_pmc_md_source_l[0]
            else:
                self.search_source = colrev.settings.SearchSource(
                    endpoint=self.endpoint,
                    filename=self.__europe_pmc_md_filename,
                    search_type=colrev.settings.SearchType.OTHER,
                    search_parameters={},
                    comment="",
                )

            self.europe_pmc_lock = Lock()

    # @classmethod
    # def check_status(cls, *, prep_operation: colrev.ops.prep.Prep) -> None:
    # ...

    @classmethod
    def __get_string_from_item(cls, *, item, key: str) -> str:  # type: ignore
        return_string = ""
        for selected_node in item.findall(key):
            return_string = selected_node.text
        return return_string

    @classmethod
    def __europe_pmc_xml_to_record(cls, *, item: Element) -> colrev.record.PrepRecord:
        retrieved_record_dict: dict = {"ENTRYTYPE": "article"}

        retrieved_record_dict.update(
            author=cls.__get_string_from_item(item=item, key="authorString")
        )
        retrieved_record_dict.update(
            journal=cls.__get_string_from_item(item=item, key="journalTitle")
        )
        retrieved_record_dict.update(
            doi=cls.__get_string_from_item(item=item, key="doi")
        )
        retrieved_record_dict.update(
            title=cls.__get_string_from_item(item=item, key="title")
        )
        retrieved_record_dict.update(
            title=cls.__get_string_from_item(item=item, key="title")
        )
        retrieved_record_dict.update(
            year=cls.__get_string_from_item(item=item, key="pubYear")
        )
        retrieved_record_dict.update(
            volume=cls.__get_string_from_item(item=item, key="journalVolume")
        )

        retrieved_record_dict.update(
            number=cls.__get_string_from_item(item=item, key="issue")
        )
        retrieved_record_dict.update(
            pmid=cls.__get_string_from_item(item=item, key="pmid")
        )
        retrieved_record_dict.update(
            epmc_source=cls.__get_string_from_item(item=item, key="source")
        )

        retrieved_record_dict.update(
            epmc_id=cls.__get_string_from_item(item=item, key="id")
        )

        retrieved_record_dict["europe_pmc_id"] = (
            retrieved_record_dict.get("epmc_source", "NO_SOURCE")
            + "/"
            + retrieved_record_dict.get("epmc_id", "NO_ID")
        )

        retrieved_record_dict["ID"] = retrieved_record_dict["europe_pmc_id"]
        retrieved_record_dict = {
            k: v
            for k, v in retrieved_record_dict.items()
            if k not in ["epmc_id", "epmc_source"] and v != ""
        }

        record = colrev.record.PrepRecord(data=retrieved_record_dict)

        # https://www.ebi.ac.uk/europepmc/webservices/rest/article/MED/23245604
        source = f"{cls.__SOURCE_URL}{record.data['europe_pmc_id']}"

        record.add_provenance_all(source=source)
        return record

    @classmethod
    def __get_similarity(
        cls,
        *,
        record: colrev.record.Record,
        retrieved_record: colrev.record.Record,
    ) -> float:
        title_similarity = fuzz.partial_ratio(
            retrieved_record.data["title"].lower(),
            record.data.get("title", "").lower(),
        )
        container_similarity = fuzz.partial_ratio(
            retrieved_record.get_container_title().lower(),
            record.get_container_title().lower(),
        )

        weights = [0.6, 0.4]
        similarities = [title_similarity, container_similarity]
        similarity = sum(similarities[g] * weights[g] for g in range(len(similarities)))
        return similarity

    def __get_europe_pmc_items(self, *, url: str, timeout: int) -> list:
        _, email = self.review_manager.get_committer()
        headers = {"user-agent": f"{__name__} (mailto:{email})"}
        session = self.review_manager.get_cached_session()
        self.review_manager.logger.debug(url)
        ret = session.request("GET", url, headers=headers, timeout=timeout)
        ret.raise_for_status()
        if ret.status_code != 200:
            return []

        root = fromstring(str.encode(ret.text))
        result_list = root.findall("resultList")[0]

        self.__next_page_url = "END"
        next_page_url_node = root.find("nextPageUrl")
        if next_page_url_node is not None:
            if next_page_url_node.text is not None:
                self.__next_page_url = next_page_url_node.text

        return result_list.findall("result")

    def __europe_pmc_query(
        self,
        *,
        record_input: colrev.record.Record,
        most_similar_only: bool = True,
        timeout: int = 60,
    ) -> list:
        """Retrieve records from Europe PMC based on a query"""

        try:
            record = record_input.copy_prep_rec()

            url = (
                "https://www.ebi.ac.uk/europepmc/webservices/rest/search?query="
                + quote(record.data["title"])
            )

            record_list = []
            while True:
                result_list = self.__get_europe_pmc_items(url=url, timeout=timeout)
                most_similar, most_similar_record = 0.0, {}
                for result_item in result_list:
                    retrieved_record = self.__europe_pmc_xml_to_record(item=result_item)

                    if "title" not in retrieved_record.data:
                        continue

                    similarity = self.__get_similarity(
                        record=record, retrieved_record=retrieved_record
                    )

                    source = (
                        f"{self.__SOURCE_URL}{retrieved_record.data['europe_pmc_id']}"
                    )
                    retrieved_record.set_masterdata_complete(
                        source=source,
                        masterdata_repository=self.review_manager.settings.is_curated_repo(),
                    )

                    if not most_similar_only:
                        record_list.append(retrieved_record)

                    elif most_similar < similarity:
                        most_similar = similarity
                        most_similar_record = retrieved_record.get_data()

                url = "END"
                if not most_similar_only:
                    url = self.__next_page_url

                if url == "END":
                    break

        except (requests.exceptions.RequestException, json.decoder.JSONDecodeError):
            return []
        except OperationalError as exc:
            raise colrev_exceptions.ServiceNotAvailableException(
                "sqlite, required for requests CachedSession "
                "(possibly caused by concurrent operations)"
            ) from exc

        if most_similar_only:
            record_list = [colrev.record.PrepRecord(data=most_similar_record)]

        return record_list

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,  # pylint: disable=unused-argument
    ) -> colrev.record.Record:
        """Retrieve masterdata from Europe PMC based on similarity with the record provided"""

        # pylint: disable=too-many-branches

        # https://www.ebi.ac.uk/europepmc/webservices/rest/article/MED/23245604

        try:
            if len(record.data.get("title", "")) > 35:
                retries = 0
                while retries < prep_operation.max_retries_on_error:
                    retries += 1

                    retrieved_records = self.__europe_pmc_query(
                        record_input=record,
                        timeout=timeout,
                    )
                    if retrieved_records:
                        retrieved_record = retrieved_records.pop()
                        break

                if not retrieved_records:
                    return record
                if 0 == len(retrieved_record.data):
                    return record

                similarity = colrev.record.PrepRecord.get_retrieval_similarity(
                    record_original=record, retrieved_record_original=retrieved_record
                )

                if similarity > prep_operation.retrieval_similarity:
                    self.europe_pmc_lock.acquire(timeout=60)

                    # Note : need to reload file because the object is not shared between processes
                    europe_pmc_feed = self.search_source.get_feed(
                        review_manager=prep_operation.review_manager,
                        source_identifier=self.source_identifier,
                        update_only=False,
                    )

                    try:
                        europe_pmc_feed.set_id(record_dict=retrieved_record.data)
                    except colrev_exceptions.NotFeedIdentifiableException:
                        return record

                    europe_pmc_feed.add_record(record=retrieved_record)

                    record.merge(
                        merging_record=retrieved_record,
                        default_source=retrieved_record.data["colrev_origin"][0],
                    )

                    record.set_masterdata_complete(
                        source=retrieved_record.data["colrev_origin"][0],
                        masterdata_repository=self.review_manager.settings.is_curated_repo(),
                    )
                    record.set_status(
                        target_state=colrev.record.RecordState.md_prepared
                    )

                    europe_pmc_feed.save_feed_file()
                    self.europe_pmc_lock.release()
                    return record

        except (requests.exceptions.RequestException, colrev_exceptions.InvalidMerge):
            self.europe_pmc_lock.release()

        return record

    def __validate_source(self) -> None:
        """Validate the SearchSource (parameters etc.)"""

        source = self.search_source

        self.review_manager.logger.debug(f"Validate SearchSource {source.filename}")

        if "query" not in source.search_parameters:
            raise colrev_exceptions.InvalidQueryException(
                "Query required in search_parameters"
            )

        self.review_manager.logger.debug(f"SearchSource {source.filename} validated")

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of Europe PMC"""

        self.__validate_source()
        # https://europepmc.org/RestfulWebService

        search_operation.review_manager.logger.info(
            f"Retrieve Europe PMC: {self.search_source.search_parameters}"
        )

        europe_pmc_feed = self.search_source.get_feed(
            review_manager=search_operation.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )

        if self.search_source.is_md_source() or self.search_source.is_quasi_md_source():
            print("Not yet implemented")
            # self.__run_md_search_update(
            #     search_operation=search_operation,
            #     europe_pmc_feed=europe_pmc_feed,
            # )

        else:
            self.__run_parameter_search(
                search_operation=search_operation,
                europe_pmc_feed=europe_pmc_feed,
                rerun=rerun,
            )

    def __run_parameter_search(
        self,
        *,
        search_operation: colrev.ops.search.Search,
        europe_pmc_feed: colrev.ops.search_feed.GeneralOriginFeed,
        rerun: bool,
    ) -> None:
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-nested-blocks

        try:
            params = self.search_source.search_parameters
            url = params["query"]

            _, email = self.review_manager.get_committer()

            headers = {"user-agent": f"{__name__} (mailto:{email})"}
            session = self.review_manager.get_cached_session()

            records = search_operation.review_manager.dataset.load_records_dict()

            while url != "END":
                self.review_manager.logger.debug(url)
                ret = session.request("GET", url, headers=headers, timeout=60)
                ret.raise_for_status()
                if ret.status_code != 200:
                    # review_manager.logger.debug(
                    #     f"europe_pmc failed with status {ret.status_code}"
                    # )
                    return

                root = fromstring(str.encode(ret.text))
                result_list = root.findall("resultList")[0]

                for result_item in result_list.findall("result"):
                    retrieved_record = self.__europe_pmc_xml_to_record(item=result_item)

                    prev_record_dict_version = {}
                    if retrieved_record.data["ID"] in europe_pmc_feed.feed_records:
                        prev_record_dict_version = deepcopy(
                            europe_pmc_feed.feed_records[retrieved_record.data["ID"]]
                        )
                    if "title" not in retrieved_record.data:
                        search_operation.review_manager.logger.warning(
                            f"Skipped record: {retrieved_record.data}"
                        )
                        continue

                    source = (
                        f"{self.__SOURCE_URL}{retrieved_record.data['europe_pmc_id']}"
                    )
                    retrieved_record.set_masterdata_complete(
                        source=source,
                        masterdata_repository=self.review_manager.settings.is_curated_repo(),
                    )

                    europe_pmc_feed.set_id(record_dict=retrieved_record.data)
                    added = europe_pmc_feed.add_record(record=retrieved_record)

                    if added:
                        search_operation.review_manager.logger.info(
                            " retrieve europe_pmc_id="
                            + retrieved_record.data["europe_pmc_id"]
                        )
                    else:
                        europe_pmc_feed.update_existing_record(
                            records=records,
                            record_dict=retrieved_record.data,
                            prev_record_dict_version=prev_record_dict_version,
                            source=self.search_source,
                            update_time_variant_fields=rerun,
                        )

                url = "END"
                next_page_url_node = root.find("nextPageUrl")
                if next_page_url_node is not None:
                    if next_page_url_node.text is not None:
                        url = next_page_url_node.text

            europe_pmc_feed.print_post_run_search_infos(records=records)

        except (requests.exceptions.RequestException, json.decoder.JSONDecodeError):
            pass
        except OperationalError as exc:
            raise colrev_exceptions.ServiceNotAvailableException(
                "sqlite, required for requests CachedSession "
                "(possibly caused by concurrent operations)"
            ) from exc
        finally:
            europe_pmc_feed.save_feed_file()

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for Europe PMC"""

        result = {"confidence": 0.0}
        if "europe_pmc_id" in data:
            result["confidence"] = 1.0

        if "https://europepmc.org" in data:  # nosec
            if data.count("https://europepmc.org") >= data.count("\n@"):
                result["confidence"] = 1.0

        return result

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.search.Search, params: str) -> None:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""

        if params is None:
            operation.add_interactively(endpoint=cls.endpoint)
            return

        host = urlparse(params).hostname

        if host and host.endswith("europepmc.org"):
            params = params.replace("https://europepmc.org/search?query=", "")

            filename = operation.get_unique_filename(file_path_string="europepmc")
            params = (
                "https://www.ebi.ac.uk/europepmc/webservices/rest/search?query="
                + params
            )
            add_source = colrev.settings.SearchSource(
                endpoint=cls.endpoint,
                filename=filename,
                search_type=colrev.settings.SearchType.DB,
                search_parameters={"query": params},
                comment="",
            )
            operation.review_manager.settings.sources.append(add_source)
            return

        raise NotImplementedError

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.ops.load_utils_bib.load_bib_file(
                load_operation=load_operation, source=self.search_source
            )
            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for Europe PMC"""
        record.data["author"].rstrip(".")
        record.data["title"].rstrip(".")
        return record



#! /usr/bin/env python
"""SearchSource: directory containing video files"""
from __future__ import annotations

from dataclasses import dataclass
from multiprocessing import Lock
from pathlib import Path

import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.built_in.search_sources.website as website_connector
import colrev.ops.load_utils_bib
import colrev.ops.search
import colrev.record
import colrev.ui_cli.cli_colors as colors

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class VideoDirSearchSource(JsonSchemaMixin):
    """SearchSource for directory containing video files"""

    # pylint: disable=too-many-instance-attributes

    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "file"
    search_type = colrev.settings.SearchType.OTHER
    api_search_supported = True
    ci_supported: bool = False
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.supported
    short_name = "Video directory"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/ops/built_in/search_sources/video_dir.md"
    )

    def __init__(
        self, *, source_operation: colrev.operation.Operation, settings: dict
    ) -> None:
        self.search_source = from_dict(data_class=self.settings_class, data=settings)
        self.source_operation = source_operation
        self.pdf_preparation_operation = (
            source_operation.review_manager.get_pdf_prep_operation(
                notify_state_transition_operation=False
            )
        )

        self.video_path = source_operation.review_manager.path / Path(
            self.search_source.search_parameters["scope"]["path"]
        )
        self.review_manager = source_operation.review_manager
        self.prep_operation = self.review_manager.get_prep_operation()
        self.url_connector = website_connector.WebsiteConnector(
            review_manager=self.review_manager
        )
        self.zotero_lock = Lock()

    def __index_video(self, *, path: Path) -> dict:
        record_dict = {"ENTRYTYPE": "online", "file": path}
        return record_dict

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of a directory containing videos"""

        search_operation.review_manager.logger.info(
            f"{colors.ORANGE}For better metadata, please add the url "
            f"(or authors and title){colors.END}"
        )

        video_feed = self.search_source.get_feed(
            review_manager=search_operation.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )

        overall_files = [
            x.relative_to(search_operation.review_manager.path)
            for x in self.video_path.glob("**/*.mp4")
        ]

        new_records_added = 0
        for file_to_add in overall_files:
            new_record = self.__index_video(path=file_to_add)

            try:
                video_feed.set_id(record_dict=new_record)
            except colrev_exceptions.NotFeedIdentifiableException:
                continue

            added = video_feed.add_record(
                record=colrev.record.Record(data=new_record),
            )
            if added:
                new_records_added += 1

        video_feed.save_feed_file()

        search_operation.review_manager.logger.info(
            f"New videos added: {new_records_added}"
        )

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for video directories"""

        result = {"confidence": 0.0}
        if filename.suffix in [".mp4"]:
            result["confidence"] = 1.0
            return result
        return result

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.search.Search, params: str) -> None:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""
        raise NotImplementedError

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Not implemented"""
        return record

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.ops.load_utils_bib.load_bib_file(
                load_operation=load_operation, source=self.search_source
            )
            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for video directories"""

        if "url" in record.data:
            self.zotero_lock = Lock()
            url_record = record.copy_prep_rec()
            self.url_connector.retrieve_md_from_website(record=url_record)
            if url_record.data.get("author", "") != "":
                record.update_field(
                    key="author", value=url_record.data["author"], source="website"
                )
            if url_record.data.get("title", "") != "":
                record.update_field(
                    key="title", value=url_record.data["title"], source="website"
                )
            self.zotero_lock.release()

        return record



#! /usr/bin/env python
"""Dedupe functionality dedicated to curated metadata repositories"""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin
from thefuzz import fuzz
from tqdm import tqdm

import colrev.env.package_manager
import colrev.ops.built_in.pdf_prep.metadata_validation
import colrev.record
import colrev.ui_cli.cli_colors as colors

if TYPE_CHECKING:
    import colrev.ops.dedupe

# pylint: disable=too-many-arguments
# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.env.package_manager.DedupePackageEndpointInterface)
@dataclass
class CurationDedupe(JsonSchemaMixin):
    """Deduplication endpoint for curations with full journals/proceedings
    retrieved from different sources (identifying duplicates in groups of
    volumes/issues or years)"""

    settings: CurationDedupeSettings
    ci_supported: bool = True

    @dataclass
    class CurationDedupeSettings(
        colrev.env.package_manager.DefaultSettings, JsonSchemaMixin
    ):
        """Settings for CurationDedupe"""

        endpoint: str
        selected_source: str

        _details = {
            "selected_source": {"tooltip": "Source (path) selected for the dedupe run"},
        }

    settings_class = CurationDedupeSettings

    def __init__(
        self,
        *,
        dedupe_operation: colrev.ops.dedupe.Dedupe,  # pylint: disable=unused-argument
        settings: dict,
    ):
        self.settings = self.settings_class.load_settings(data=settings)

        pdf_prep_operation = dedupe_operation.review_manager.get_pdf_prep_operation()
        self.pdf_metadata_validation = (
            colrev.ops.built_in.pdf_prep.metadata_validation.PDFMetadataValidation(
                pdf_prep_operation=pdf_prep_operation,
                settings={"endpoint": "dedupe_pdf_md_validation"},
            )
        )

    def __get_similarity(self, *, df_a: pd.Series, df_b: pd.Series) -> float:
        author_similarity = fuzz.ratio(df_a["author"], df_b["author"]) / 100

        title_similarity = (
            fuzz.ratio(df_a["title"].lower(), df_b["title"].lower()) / 100
        )

        # Note : the toc-based processing means that we are robust against
        # outlet, year, volume, number variations!
        weights = [0.4, 0.6]
        similarities = [
            author_similarity,
            title_similarity,
        ]

        weighted_average = sum(
            similarities[g] * weights[g] for g in range(len(similarities))
        )

        similarity_score = round(weighted_average, 4)

        return similarity_score

    def __calculate_similarities(
        self,
        *,
        similarity_array: np.ndarray,
        references: pd.DataFrame,
        min_similarity: float,
    ) -> tuple:
        # Fill out the similarity matrix first
        for base_entry_i in range(1, references.shape[0]):
            for comparison_entry_i in range(1, references.shape[0]):
                if base_entry_i > comparison_entry_i:
                    if -1 != similarity_array[base_entry_i, comparison_entry_i]:
                        similarity_array[
                            base_entry_i, comparison_entry_i
                        ] = self.__get_similarity(
                            df_a=references.iloc[base_entry_i],
                            df_b=references.iloc[comparison_entry_i],
                        )

        tuples_to_process = []
        maximum_similarity = 1
        while True:
            maximum_similarity = np.amax(similarity_array)
            if maximum_similarity < min_similarity:
                break
            result = np.where(similarity_array == np.amax(similarity_array))
            list_of_coordinates = list(zip(result[0], result[1]))
            for cord in list_of_coordinates:
                similarity_array[cord] = 0  # ie., has been processed
                tuples_to_process.append(
                    [
                        references.iloc[cord[0]]["ID"],
                        references.iloc[cord[1]]["ID"],
                        maximum_similarity,
                        "not_processed",
                    ]
                )

        return similarity_array, tuples_to_process

    def __get_toc_items(self, *, records_list: list) -> list:
        toc_items = []
        for record in records_list:
            toc_item = {}
            if record["ENTRYTYPE"] == "article":
                if "journal" in record:
                    toc_item["journal"] = record["journal"]
                if "volume" in record:
                    toc_item["volume"] = record["volume"]
                if "number" in record:
                    toc_item["number"] = record["number"]

            if record["ENTRYTYPE"] == "inproceedings":
                if "booktitle" in record:
                    toc_item["booktitle"] = record["booktitle"]
                    toc_item["year"] = record["year"]
            if len(toc_item) > 0:
                toc_items.append(toc_item)

        temp = {tuple(sorted(sub.items())) for sub in toc_items}
        toc_items = list(map(dict, temp))  # type: ignore
        return toc_items

    def __warn_on_missing_sources(
        self, *, dedupe_operation: colrev.ops.dedupe.Dedupe, first_source: bool
    ) -> None:
        # warn if not all SOURCE.filenames are included in a dedupe script
        if first_source:
            available_sources = [
                str(s.filename)
                for s in dedupe_operation.review_manager.settings.sources
                if "md_" not in str(s.filename)
            ]
            dedupe_sources = [
                s["selected_source"]
                for s in dedupe_operation.review_manager.settings.dedupe.dedupe_package_endpoints
                if "colrev.curation_full_outlet_dedupe" == s["endpoint"]
            ]
            sources_missing_in_dedupe = [
                x for x in available_sources if x not in dedupe_sources
            ]
            if len(sources_missing_in_dedupe) > 0:
                dedupe_operation.review_manager.logger.warning(
                    f"{colors.ORANGE}Sources missing in "
                    "dedupe.scripts.colrev.curation_full_outlet_dedupe: "
                    f"{','.join(sources_missing_in_dedupe)}{colors.END}"
                )
                if input("Add sources [y,n]?") == "y":
                    for source_missing_in_dedupe in sources_missing_in_dedupe:
                        dedupe_package_endpoints = (
                            dedupe_operation.review_manager.settings.dedupe.dedupe_package_endpoints
                        )
                        penultimate_position = len(dedupe_package_endpoints) - 1
                        dedupe_script_to_add = {
                            "endpoint": "colrev.curation_full_outlet_dedupe",
                            "selected_source": source_missing_in_dedupe,
                        }
                        dedupe_package_endpoints.insert(
                            penultimate_position, dedupe_script_to_add
                        )
                        dedupe_operation.review_manager.save_settings()
                        dedupe_operation.review_manager.logger.info(
                            f"{colors.GREEN}Added {source_missing_in_dedupe} "
                            f"to dedupe.scripts{colors.END}"
                        )

    def __add_first_source_if_deduplicated(
        self, *, dedupe_operation: colrev.ops.dedupe.Dedupe, records: dict
    ) -> None:
        dedupe_operation.review_manager.logger.info(
            f"Starting with records from {self.settings.selected_source}"
            " (setting to md_processed as the initial records)"
        )

        source_records = [
            r
            for r in records.values()
            if r["colrev_status"] == colrev.record.RecordState.md_prepared
            and any(
                self.settings.selected_source.replace("data/search/", "") in co
                for co in r["colrev_origin"]
            )
        ]

        toc_items = self.__get_toc_items(records_list=source_records)

        for toc_item in toc_items:
            # Note : these would be potential errors (duplicates)
            # because they have the same selected_source
            processed_same_toc_same_source_records = [
                r
                for r in records.values()
                if all(r.get(k, "NA") == v for k, v in toc_item.items())
                and r["colrev_status"]
                not in [
                    colrev.record.RecordState.md_prepared,
                    colrev.record.RecordState.md_needs_manual_preparation,
                    colrev.record.RecordState.md_imported,
                    colrev.record.RecordState.rev_prescreen_excluded,
                ]
                and any(
                    self.settings.selected_source.replace("data/search/", "") in co
                    for co in r["colrev_origin"]
                )
            ]
            if 0 == len(processed_same_toc_same_source_records):
                print("\n\n")
                print(toc_item)

                for source_record_dict in sorted(
                    source_records, key=lambda d: d["author"]
                ):
                    if all(
                        source_record_dict.get(k, "NA") == v
                        for k, v in toc_item.items()
                    ):
                        # Record(data=sr).print_citation_format()
                        print(
                            f"{source_record_dict.get('author', 'NO_AUTHOR')} : "
                            f"{source_record_dict.get('title', 'NO_TITLE')}"
                        )
                recs_unique = dedupe_operation.review_manager.force_mode
                if not recs_unique:
                    recs_unique = "y" == input(
                        "No existing records (md_processed*) found. "
                        "All records unique? Set to md_processed [y]? "
                    )
                if recs_unique:
                    for source_record_dict in source_records:
                        if all(
                            source_record_dict.get(k, "NA") == v
                            for k, v in toc_item.items()
                        ):
                            source_record = colrev.record.Record(
                                data=source_record_dict
                            )
                            source_record.set_status(
                                target_state=colrev.record.RecordState.md_processed
                            )
            else:
                print(toc_item)
                print("Pre-imported records found for this toc_item (skipping)")
                # print(processed_same_toc_same_source_records)

        for record in records.values():
            record.pop("container_title")
        dedupe_operation.review_manager.dataset.save_records_dict(records=records)
        dedupe_operation.review_manager.dataset.add_record_changes()

        if dedupe_operation.review_manager.dataset.has_changes():
            dedupe_operation.review_manager.logger.info(
                f"{colors.GREEN}Commit changes{colors.END}"
            )
            dedupe_operation.review_manager.create_commit(
                msg=(
                    "Merge duplicate records (set unique records from "
                    f"{self.settings.selected_source} "
                    "to md_processed)"
                ),
            )
        else:
            dedupe_operation.review_manager.logger.info(
                f"{colors.GREEN}No duplicates found{colors.END}"
            )

    def __prep_records(self, *, records: dict) -> dict:
        required_fields = [
            "title",
            "author",
            "year",
            "journal",
            "volume",
            "number",
            "pages",
            "booktitle",
        ]
        for record in records.values():
            if "container_title" not in record:
                record["container_title"] = (
                    record.get("journal", "")
                    + record.get("booktitle", "")
                    + record.get("series", "")
                )

            for required_field in required_fields:
                if required_field not in record:
                    record[required_field] = ""
        return records

    def __dedupe_source(
        self, *, dedupe_operation: colrev.ops.dedupe.Dedupe, records: dict
    ) -> list[dict]:
        dedupe_operation.review_manager.logger.info(
            "Processing as a non-pdf source (matching exact colrev_ids)"
        )

        source_records = [
            r
            for r in records.values()
            if r["colrev_status"] == colrev.record.RecordState.md_prepared
            and any(
                self.settings.selected_source.replace("data/search/", "") in co
                for co in r["colrev_origin"]
            )
        ]

        toc_items = self.__get_toc_items(records_list=source_records)

        decision_list: list[dict] = []
        # decision_list =[{'ID1': ID1, 'ID2': ID2, 'decision': 'duplicate'}]

        # match based on overlapping  colrev_ids
        for toc_item in tqdm(toc_items):
            processed_same_toc_records = [
                r
                for r in records.values()
                if all(r.get(k, "NA") == v for k, v in toc_item.items())
                and r["colrev_status"]
                not in [
                    colrev.record.RecordState.md_imported,
                    colrev.record.RecordState.md_needs_manual_preparation,
                    colrev.record.RecordState.md_prepared,
                    colrev.record.RecordState.rev_prescreen_excluded,
                ]
                and not any(
                    self.settings.selected_source.replace("data/search/", "") in co
                    for co in r["colrev_origin"]
                )
            ]
            new_same_toc_records = [
                r
                for r in source_records
                if all(r.get(k, "NA") == v for k, v in toc_item.items())
            ]
            if len(new_same_toc_records) > 0:
                # print(new_same_toc_records)
                for new_same_toc_record in new_same_toc_records:
                    for rec2 in processed_same_toc_records:
                        overlapping_colrev_ids = colrev.record.Record(
                            data=new_same_toc_record
                        ).has_overlapping_colrev_id(
                            record=colrev.record.Record(data=rec2)
                        )
                        if overlapping_colrev_ids:
                            decision_list.append(
                                {
                                    "ID1": new_same_toc_record["ID"],
                                    "ID2": rec2["ID"],
                                    "decision": "duplicate",
                                }
                            )
                            print("TODO : validate whether it merges correctly:")
                            input(decision_list)

        return decision_list

    def __validate_potential_merge(
        self, *, rec1: dict, rec2: dict, dedupe_operation: colrev.ops.dedupe.Dedupe
    ) -> dict:
        if "file" in rec2:
            updated_record = rec1.copy()
            updated_record["file"] = rec2["file"]
        elif "file" in rec1:
            updated_record = rec2.copy()
            updated_record["file"] = rec1["file"]
        else:  # None of the records is curated
            raise FileNotFoundError

        record = colrev.record.Record(data=updated_record)
        validation_info = self.pdf_metadata_validation.validates_based_on_metadata(
            review_manager=dedupe_operation.review_manager,
            record=record,
        )
        return validation_info

    def __process_pdf_tuple(
        self,
        *,
        tuple_to_process: tuple,
        records: dict,
        decision_list: list,
        curated_record_ids: list,
        pdf_record_ids: list,
        dedupe_operation: colrev.ops.dedupe.Dedupe,
    ) -> None:
        rec1 = records[tuple_to_process[0]]
        rec2 = records[tuple_to_process[1]]

        # Note : Focus on merges between
        # curated_records and pdf_same_toc_records
        # Note : this should also ensure that pdf groups are not merged
        # until a corresponding curated record group is available.
        if (rec1["ID"] in curated_record_ids and rec2["ID"] in curated_record_ids) or (
            rec1["ID"] in pdf_record_ids and rec2["ID"] in pdf_record_ids
        ):
            return

        try:
            validation_info = self.__validate_potential_merge(
                rec1=rec1, rec2=rec2, dedupe_operation=dedupe_operation
            )
        except FileNotFoundError:
            return

        overlapping_colrev_ids = colrev.record.Record(
            data=rec1
        ).has_overlapping_colrev_id(record=colrev.record.Record(data=rec2))
        if validation_info["validates"] or overlapping_colrev_ids:
            # Note : make sure that we merge into the CURATED record
            if "file" in rec1:
                if tuple_to_process[0] in [x["ID1"] for x in decision_list]:
                    return
                if rec1["colrev_status"] < rec2["colrev_status"]:
                    decision_list.append(
                        {
                            "ID1": tuple_to_process[1],
                            "ID2": tuple_to_process[0],
                            "decision": "duplicate",
                        }
                    )
                else:
                    decision_list.append(
                        {
                            "ID1": tuple_to_process[0],
                            "ID2": tuple_to_process[1],
                            "decision": "duplicate",
                        }
                    )
            else:
                if tuple_to_process[1] in [x["ID1"] for x in decision_list]:
                    return
                if rec1["colrev_status"] < rec2["colrev_status"]:
                    decision_list.append(
                        {
                            "ID1": tuple_to_process[1],
                            "ID2": tuple_to_process[0],
                            "decision": "duplicate",
                        }
                    )
                else:
                    decision_list.append(
                        {
                            "ID1": tuple_to_process[0],
                            "ID2": tuple_to_process[1],
                            "decision": "duplicate",
                        }
                    )

    def __dedupe_pdf_toc_item(
        self,
        *,
        decision_list: list[dict],
        toc_item: dict,
        records: dict,
        source_records: list,
        dedupe_operation: colrev.ops.dedupe.Dedupe,
    ) -> None:
        processed_same_toc_records = [
            r
            for r in records.values()
            if all(r.get(k, "NA") == v for k, v in toc_item.items())
            and r["colrev_status"]
            not in [
                colrev.record.RecordState.md_imported,
                colrev.record.RecordState.md_needs_manual_preparation,
                colrev.record.RecordState.md_prepared,
                colrev.record.RecordState.rev_prescreen_excluded,
            ]
            and not any(
                self.settings.selected_source.replace("data/search/", "") in co
                for co in r["colrev_origin"]
            )
        ]
        pdf_same_toc_records = [
            r
            for r in source_records
            if all(r.get(k, "NA") == v for k, v in toc_item.items())
        ]

        references = pd.DataFrame.from_records(
            processed_same_toc_records + pdf_same_toc_records
        )

        nr_entries = references.shape[0]
        if nr_entries == 0:
            return
        similarity_array = np.zeros([nr_entries, nr_entries])

        # Note : min_similarity only means that the PDF will be considered
        # for validates_based_on_metadata(...), which is the acutal test!
        similarity_array, tuples_to_process = self.__calculate_similarities(
            similarity_array=similarity_array,
            references=references,
            min_similarity=0.7,
        )

        curated_record_ids = [r["ID"] for r in processed_same_toc_records]
        pdf_record_ids = [r["ID"] for r in pdf_same_toc_records]
        for tuple_to_process in tuples_to_process:
            self.__process_pdf_tuple(
                tuple_to_process=tuple_to_process,
                records=records,
                decision_list=decision_list,
                curated_record_ids=curated_record_ids,
                pdf_record_ids=pdf_record_ids,
                dedupe_operation=dedupe_operation,
            )

    def __dedupe_pdf_source(
        self, *, dedupe_operation: colrev.ops.dedupe.Dedupe, records: dict
    ) -> list[dict]:
        dedupe_operation.review_manager.logger.info("Processing as a pdf source")

        source_records = [
            r
            for r in records.values()
            if r["colrev_status"] == colrev.record.RecordState.md_prepared
            and any(
                self.settings.selected_source.replace("data/search/", "") in co
                for co in r["colrev_origin"]
            )
        ]

        decision_list: list[dict] = []
        # decision_list =[{'ID1': ID1, 'ID2': ID2, 'decision': 'duplicate'}]

        for toc_item in tqdm(self.__get_toc_items(records_list=source_records)):
            self.__dedupe_pdf_toc_item(
                decision_list=decision_list,
                toc_item=toc_item,
                records=records,
                source_records=source_records,
                dedupe_operation=dedupe_operation,
            )

        return decision_list

    def __pdf_source_selected(
        self, *, dedupe_operation: colrev.ops.dedupe.Dedupe
    ) -> bool:
        pdf_source = False
        relevant_source = [
            s
            for s in dedupe_operation.review_manager.settings.sources
            if str(s.filename) == self.settings.selected_source
        ]
        if len(relevant_source) > 0:
            pdf_source = "colrev.pdfs_dir" == relevant_source[0].endpoint
        return pdf_source

    def __first_source_selected(
        self, *, dedupe_operation: colrev.ops.dedupe.Dedupe
    ) -> bool:
        return (
            self.settings.selected_source
            == dedupe_operation.review_manager.settings.dedupe.dedupe_package_endpoints[
                0
            ]["selected_source"]
        )

    def run_dedupe(self, dedupe_operation: colrev.ops.dedupe.Dedupe) -> None:
        """Run the curation dedupe procedure"""

        records = dedupe_operation.review_manager.dataset.load_records_dict()
        records = self.__prep_records(records=records)

        # first_source should be the highest quality source
        # (which moves to md_processed first)
        first_source = self.__first_source_selected(dedupe_operation=dedupe_operation)

        self.__warn_on_missing_sources(
            dedupe_operation=dedupe_operation, first_source=first_source
        )

        if first_source:
            self.__add_first_source_if_deduplicated(
                dedupe_operation=dedupe_operation, records=records
            )
            return

        dedupe_operation.review_manager.logger.info(
            "Identify duplicates between "
            f"curated_records and {self.settings.selected_source} (within toc_items)"
        )

        decision_list: list[dict] = []
        # decision_list =[{'ID1': ID1, 'ID2': ID2, 'decision': 'duplicate'}]
        if not self.__pdf_source_selected(dedupe_operation=dedupe_operation):
            decision_list = self.__dedupe_source(
                dedupe_operation=dedupe_operation, records=records
            )
        else:
            decision_list = self.__dedupe_pdf_source(
                dedupe_operation=dedupe_operation, records=records
            )

        # Note : dedupe.apply_merges reloads the records and
        # thereby discards previous changes
        if len(decision_list) == 0:
            dedupe_operation.review_manager.logger.info(
                f"{colors.GREEN}No merge-candidates identified between sets{colors.END}"
            )
            return

        dedupe_operation.review_manager.logger.info(
            f"{colors.GREEN}Duplicates identified{colors.END}"
        )

        # decision_list =
        # [{'ID1': '0000000053', 'ID2': 'BellMillsFadel2013', 'decision': 'duplicate'}, .. . ]

        preferred_masterdata_sources = [
            s
            for s in dedupe_operation.review_manager.settings.sources
            if s.endpoint != "colrev.pdfs_dir"
        ]
        dedupe_operation.apply_merges(
            results=decision_list,
            preferred_masterdata_sources=preferred_masterdata_sources,
        )

        dedupe_operation.review_manager.dataset.add_record_changes()

        dedupe_operation.review_manager.create_commit(
            msg="Merge duplicate records",
        )



#! /usr/bin/env python
"""Convenience functions to load ris files (based on rispy)"""
from __future__ import annotations

import re
from pathlib import Path
from typing import TYPE_CHECKING

import rispy
from rispy import BaseParser
from rispy.config import LIST_TYPE_TAGS
from rispy.config import TAG_KEY_MAPPING

if TYPE_CHECKING:
    import colrev.ops.load


# Based on https://github.com/aurimasv/translators/wiki/RIS-Tag-Map
REFERENCE_TYPES = {
    "JOUR": "article",
    "JFULL": "article",
    "ABST": "article",
    "INPR": "article",  # inpress
    "CONF": "inproceedings",
    "CPAPER": "inproceedings",
    "THES": "phdthesis",
    "REPT": "techreport",
    "RPRT": "techreport",
    "CHAP": "inbook",
    "BOOK": "book",
}
KEY_MAP = {
    "article": {
        "ID": "ID",
        "ENTRYTYPE": "ENTRYTYPE",
        "year": "year",
        "authors": "author",
        "primary_title": "title",
        "secondary_title": "journal",
        "notes_abstract": "abstract",
        "volume": "volume",
        "number": "number",
        "doi": "doi",
        "publisher": "publisher",
        "url": "url",
        "fulltext": "fulltext",
        "pubmedid": "pubmedid",
        "keywords": "keywords",
        "pages": "pages",
    },
    "inproceedings": {
        "ID": "ID",
        "ENTRYTYPE": "ENTRYTYPE",
        "year": "year",
        "authors": "author",
        "primary_title": "title",
        "secondary_title": "booktitle",
        "doi": "doi",
        "url": "url",
        "fulltext": "fulltext",
        "pubmedid": "pubmedid",
        "keywords": "keywords",
        "pages": "pages",
    },
    "inbook": {
        "ID": "ID",
        "ENTRYTYPE": "ENTRYTYPE",
        "year": "year",
        "authors": "author",
        "primary_title": "chapter",
        "secondary_title": "title",
        "doi": "doi",
        "publisher": "publisher",
        "edition": "edition",
        "url": "url",
        "fulltext": "fulltext",
        "keywords": "keywords",
        "pages": "pages",
    },
    "techreport": {
        "ID": "ID",
        "ENTRYTYPE": "ENTRYTYPE",
        "year": "year",
        "authors": "author",
        "primary_title": "title",
        "url": "url",
        "fulltext": "fulltext",
        "keywords": "keywords",
        "publisher": "publisher",
        "pages": "pages",
    },
}


class DefaultRISParser(BaseParser):
    """Default parser for RIS files."""

    START_TAG = "TY"
    IGNORE = ["FN", "VR", "EF"]
    PATTERN = r"^[A-Z][A-Z0-9]+ |^ER\s?|^EF\s?"
    DEFAULT_MAPPING = TAG_KEY_MAPPING
    DEFAULT_LIST_TAGS = LIST_TYPE_TAGS + ["UR"]

    def get_content(self, line: str) -> str:
        "Get the content from a line."
        return line[line.find(" - ") + 2 :].strip()

    def is_header(self, line: str) -> bool:
        "Check whether the line is a header element"
        return not re.match("[A-Z0-9]+  - ", line)


class RISLoader:

    """Loads ris files"""

    def __init__(
        self,
        *,
        load_operation: colrev.ops.load.Load,
        source: colrev.settings.SearchSource,
    ):
        self.load_operation = load_operation
        self.source = source

    def apply_ris_fixes(self, *, filename: Path) -> None:
        """Fix common defects in RIS files"""

        # Error to fix: for lists of keywords, each line should start with the KW tag

        with open(filename, encoding="UTF-8") as file:
            lines = [line.rstrip("\n") for line in file]
            for i, line in enumerate(lines):
                if line.startswith("PMID "):
                    lines[i] = line.replace("PMID ", "PM ")

            # add missing start tags in lists (like KW)
            processing_tag = ""
            for i, line in enumerate(lines):
                tag_match = re.match(r"^[A-Z][A-Z0-9]+(\s+)-", line)  # |^ER\s?|^EF\s?
                if tag_match:
                    processing_tag = tag_match.group()
                elif line == "":
                    processing_tag = ""
                    continue
                elif processing_tag == "":
                    continue
                else:
                    lines[i] = f"{processing_tag} {line}"

        with open(filename, "w", encoding="utf-8") as file:
            for line in lines:
                file.write(f"{line}\n")

    def load_ris_entries(
        self, *, filename: Path, ris_parser: BaseParser = DefaultRISParser
    ) -> dict:
        """Load ris entries

        The resulting keys should coincide with those in the KEY_MAP
        but they can be adapted before calling the convert_to_records()"""

        # Note : depending on the source, a specific ris_parser implementation may be selected.
        # its DEFAULT_LIST_TAGS can be extended with list fiels that should be joined automatically

        self.load_operation.ensure_append_only(file=self.source.filename)

        with open(filename, encoding="utf-8") as ris_file:
            entries = rispy.load(file=ris_file, implementation=ris_parser)

        for entry in entries:
            if "pages" in entry:
                continue
            if "start_page" in entry and "end_page" in entry:
                entry["pages"] = f"{entry.pop('start_page')}--{entry.pop('end_page')}"
            elif "start_page" in entry:
                entry["pages"] = f"{entry.pop('start_page')}"

            for key in [
                ris_parser.DEFAULT_MAPPING[k] for k in ris_parser.DEFAULT_LIST_TAGS
            ]:
                if key not in entry:
                    continue
                if "author" in key:
                    entry[key] = " and ".join(entry[key])
                elif "url" == key:
                    urls = entry["url"]
                    for url in urls:
                        if url.endswith(".pdf"):
                            entry["fulltext"] = url
                        else:
                            entry["url"] = url
                            break
                else:
                    entry[key] = ", ".join(entry[key])

        return entries

    def convert_to_records(self, *, entries: dict) -> dict:
        """Converts ris entries it to bib records"""

        # Note : REFERENCE_TYPES and KEY_MAP are hard-coded (standard)
        # This function intentionally fails when the input does not comply
        # with this standard

        records: dict = {}
        for counter, entry in enumerate(entries):
            _id = str(counter + 1).zfill(5)

            type_of_ref = entry["type_of_reference"]
            if type_of_ref not in REFERENCE_TYPES:
                raise NotImplementedError(f"Undefined reference type: {type_of_ref}")

            entry_type = REFERENCE_TYPES[type_of_ref]
            entry["ID"] = _id
            entry["ENTRYTYPE"] = entry_type

            if entry_type not in KEY_MAP:
                raise NotImplementedError(f"No KEY_MAP defined for {entry_type}")

            record: dict = {}
            for ris_key, key in KEY_MAP[entry_type].items():
                if ris_key in entry:
                    record[key] = entry[ris_key]

            records[_id] = record

        return records



#! /usr/bin/env python
"""SearchSource: OpenAlex"""
from __future__ import annotations

from dataclasses import dataclass
from multiprocessing import Lock
from pathlib import Path
from typing import Optional
from typing import TYPE_CHECKING

import pyalex
import requests
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin
from pyalex import Works

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.load_utils_bib
import colrev.record

if TYPE_CHECKING:
    import colrev.ops.search
    import colrev.ops.prep

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class OpenAlexSearchSource(JsonSchemaMixin):
    """SearchSource for the OpenAlex API"""

    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "openalex_id"
    # "https://api.crossref.org/works/{{doi}}"
    search_type = colrev.settings.SearchType.DB
    api_search_supported = True
    ci_supported: bool = True
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.oni
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/ops/built_in/search_sources/open_alex.md"
    )
    short_name = "OpenAlex"
    __open_alex_md_filename = Path("data/search/md_open_alex.bib")

    def __init__(
        self,
        *,
        source_operation: colrev.operation.Operation,
        settings: Optional[dict] = None,
    ) -> None:
        # Note: not yet implemented
        # Note : once this is implemented, add "colrev.open_alex_prep" to the default settings
        # if settings:
        #     # OpenAlex as a search_source
        #     self.search_source = from_dict(
        #         data_class=self.settings_class, data=settings
        #     )
        # else:
        # OpenAlex as an md-prep source
        open_alex_md_source_l = [
            s
            for s in source_operation.review_manager.settings.sources
            if s.filename == self.__open_alex_md_filename
        ]
        if open_alex_md_source_l:
            self.search_source = open_alex_md_source_l[0]
        else:
            self.search_source = colrev.settings.SearchSource(
                endpoint="colrev.open_alex",
                filename=self.__open_alex_md_filename,
                search_type=colrev.settings.SearchType.OTHER,
                search_parameters={},
                comment="",
            )

        self.open_alex_lock = Lock()

        self.language_service = colrev.env.language_service.LanguageService()

        self.review_manager = source_operation.review_manager
        _, pyalex.config.email = self.review_manager.get_committer()

    def check_availability(
        self, *, source_operation: colrev.operation.Operation
    ) -> None:
        """Check status (availability) of the OpenAlex API"""

    def __set_author_from_item(self, *, record_dict: dict, item: dict) -> None:
        ret_str = ""
        for author in item["authorships"]:
            if "author" not in author:
                continue
            if author["author"].get("display_name", None) is None:
                continue
            author_string = colrev.record.PrepRecord.format_author_field(
                input_string=author["author"]["display_name"]
            )
            if ret_str == "":
                ret_str = author_string
            else:
                ret_str += " and " + author_string
        if ret_str != "":
            record_dict["author"] = ret_str

    def __parse_item_to_record(self, *, item: dict) -> colrev.record.Record:
        def set_entrytype(*, record_dict: dict, item: dict) -> None:
            if "title" in record_dict and record_dict["title"] is None:
                del record_dict["title"]
            if item.get("type_crossref", "") == "proceedings-article":
                record_dict["ENTRYTYPE"] = "inproceedings"
                if (
                    item.get("primary_location", None) is not None
                    and item["primary_location"].get("source", None) is not None
                ):
                    display_name = item["primary_location"]["source"]["display_name"]
                    if display_name != "Proceedings":
                        record_dict["booktitle"] = display_name
            elif item["type"] in ["journal-article", "article"]:
                record_dict["ENTRYTYPE"] = "article"
                if (
                    item.get("primary_location", None) is not None
                    and item["primary_location"].get("source", None) is not None
                ):
                    record_dict["journal"] = item["primary_location"]["source"][
                        "display_name"
                    ]
            else:
                record_dict["ENTRYTYPE"] = "misc"

        record_dict = {}
        record_dict["id"] = item["id"].replace("https://openalex.org/", "")
        if "title" in item and item["title"] is not None:
            record_dict["title"] = item["title"].lstrip("[").rstrip("].")
        set_entrytype(record_dict=record_dict, item=item)

        if "publication_year" in item and item["publication_year"] is not None:
            record_dict["year"] = str(item["publication_year"])
        if "language" in item and item["language"] is not None:
            record_dict["language"] = item["language"]

        if "is_retracted" in item and item["is_retracted"]:
            record_dict["retracted"] = item["is_retracted"]

        if "doi" in item and item["doi"] is not None:
            record_dict["doi"] = item["doi"].upper().replace("HTTPS://DOI.ORG/", "")

        record_dict["cited_by"] = item["cited_by_count"]

        if "volume" in item["biblio"] and item["biblio"]["volume"] is not None:
            record_dict["volume"] = item["biblio"]["volume"]
        if "issue" in item["biblio"] and item["biblio"]["issue"] is not None:
            record_dict["number"] = item["biblio"]["issue"]
        if "first_page" in item["biblio"] and item["biblio"]["first_page"] is not None:
            record_dict["pages"] = item["biblio"]["first_page"]
        if "last_page" in item["biblio"] and item["biblio"]["last_page"] is not None:
            record_dict["pages"] += "--" + item["biblio"]["last_page"]

        self.__set_author_from_item(record_dict=record_dict, item=item)
        record = colrev.record.Record(data=record_dict)

        self.__fix_errors(record=record)
        return record

    def __fix_errors(self, *, record: colrev.record.Record) -> None:
        if "PubMed" == record.data.get("journal", ""):
            record.remove_field(key="journal")
        try:
            self.language_service.unify_to_iso_639_3_language_codes(record=record)
        except colrev_exceptions.InvalidLanguageCodeException:
            record.remove_field(key="language")

    def __get_masterdata_record(
        self, *, record: colrev.record.Record
    ) -> colrev.record.Record:
        try:
            retrieved_record = self.__parse_item_to_record(
                item=Works()[record.data["colrev.open_alex.id"]]
            )

            self.open_alex_lock.acquire(timeout=120)

            # Note : need to reload file because the object is not shared between processes
            open_alex_feed = self.search_source.get_feed(
                review_manager=self.review_manager,
                source_identifier=self.source_identifier,
                update_only=False,
            )

            open_alex_feed.set_id(record_dict=retrieved_record.data)
            open_alex_feed.add_record(record=retrieved_record)
            record.change_entrytype(
                new_entrytype=retrieved_record.data["ENTRYTYPE"],
                qm=self.review_manager.get_qm(),
            )

            record.merge(
                merging_record=retrieved_record,
                default_source=retrieved_record.data["colrev_origin"][0],
            )
            open_alex_feed.save_feed_file()
        except (
            colrev_exceptions.InvalidMerge,
            colrev_exceptions.RecordNotParsableException,
            requests.exceptions.RequestException,
        ):
            pass
        except Exception as exc:
            raise exc
        finally:
            try:
                self.open_alex_lock.release()
            except ValueError:
                pass

        return record

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 30,
    ) -> colrev.record.Record:
        """Retrieve masterdata from OpenAlex based on similarity with the record provided"""

        if "colrev.open_alex.id" not in record.data:
            # Note: not yet implemented
            # https://github.com/OpenAPC/openapc-de/blob/master/python/import_dois.py
            # if len(record.data.get("title", "")) < 35 and "doi" not in record.data:
            #     return record
            # record = self.__check_doi_masterdata(record=record)
            return record

        record = self.__get_masterdata_record(record=record)

        return record

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of OpenAlex"""

        # https://docs.openalex.org/api-entities/works

        # crossref_feed = self.search_source.get_feed(
        #     review_manager=search_operation.review_manager,
        #     source_identifier=self.source_identifier,
        #     update_only=(not rerun),
        # )

        # try:
        #     if (
        #         self.search_source.is_md_source()
        #         or self.search_source.is_quasi_md_source()
        #     ):
        #         self.__run_md_search_update(
        #             search_operation=search_operation,
        #             crossref_feed=crossref_feed,
        #         )

        #     else:
        #         self.__run_parameter_search(
        #             search_operation=search_operation,
        #             crossref_feed=crossref_feed,
        #             rerun=rerun,
        #         )
        # except (
        #     requests.exceptions.Timeout,
        #     requests.exceptions.JSONDecodeError,
        # ) as exc:
        #     # watch github issue:
        #     # https://github.com/fabiobatalha/crossrefapi/issues/46
        #     if "504 Gateway Time-out" in str(exc):
        #         raise colrev_exceptions.ServiceNotAvailableException(
        #             self.__availability_exception_message
        #         )
        #     raise colrev_exceptions.ServiceNotAvailableException(
        #         self.__availability_exception_message
        #     )

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for OpenAlex"""

        result = {"confidence": 0.0}

        return result

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.search.Search, params: str) -> None:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""

        raise colrev_exceptions.PackageParameterError(
            f"Cannot add OpenAlex endpoint with query {params}"
        )

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.ops.load_utils_bib.load_bib_file(
                load_operation=load_operation, source=self.search_source
            )
            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for OpenAlex"""

        return record



#! /usr/bin/env python
"""Completion of metadata based on year-volume-issue dependency as a prep operation"""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import requests
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.built_in.search_sources.crossref as crossref_connector
import colrev.ops.search_sources
import colrev.record

# pylint: disable=duplicate-code

if TYPE_CHECKING:
    import colrev.ops.prep

# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
@dataclass
class YearVolIssPrep(JsonSchemaMixin):
    """Prepares records based on year-volume-issue dependency"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = True

    source_correction_hint = (
        "ask the publisher to correct the metadata"
        + " (see https://www.crossref.org/blog/"
        + "metadata-corrections-updates-and-additions-in-metadata-manager/"
    )
    always_apply_changes = True

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)
        self.review_manager = prep_operation.review_manager
        self.local_index = prep_operation.review_manager.get_local_index()
        self.vol_nr_dict = self.__get_vol_nr_dict()
        self.quality_model = self.review_manager.get_qm()

    def __get_vol_nr_dict(self) -> dict:
        vol_nr_dict: dict = {}
        if not hasattr(self.review_manager, "dataset"):
            return vol_nr_dict
        records = self.review_manager.dataset.load_records_dict()
        for record in records.values():
            # pylint: disable=duplicate-code
            if record[
                "colrev_status"
            ] not in colrev.record.RecordState.get_post_x_states(
                state=colrev.record.RecordState.md_processed
            ):
                continue
            if not record.get("year", "NA").isdigit():
                continue

            if "journal" not in record or "volume" not in record:
                continue

            if record["journal"] not in vol_nr_dict:
                vol_nr_dict[record["journal"]] = {}

            if record["volume"] not in vol_nr_dict[record["journal"]]:
                vol_nr_dict[record["journal"]][record["volume"]] = {}

            if "number" not in record:
                vol_nr_dict[record["journal"]][record["volume"]] = record["year"]
            else:
                if isinstance(vol_nr_dict[record["journal"]][record["volume"]], dict):
                    vol_nr_dict[record["journal"]][record["volume"]][
                        record["number"]
                    ] = record["year"]
                else:
                    # do not use inconsistent data (has/has no number)
                    del vol_nr_dict[record["journal"]][record["volume"]]

        return vol_nr_dict

    def __get_year_from_toc(self, *, record: colrev.record.Record) -> None:
        # TBD: maybe extract the following three lines as a separate script...
        try:
            year = self.local_index.get_year_from_toc(record_dict=record.get_data())
            record.update_field(
                key="year",
                value=year,
                source="LocalIndexPrep",
                note="",
                keep_source_if_equal=True,
            )
        except colrev_exceptions.TOCNotAvailableException:
            pass

    def __get_year_from_vol_nr_dict(self, *, record: colrev.record.Record) -> None:
        if "journal" not in record.data or "volume" not in record.data:
            return

        if record.data["journal"] not in self.vol_nr_dict:
            return

        if record.data["volume"] not in self.vol_nr_dict[record.data["journal"]]:
            return

        if "number" in record.data:
            if (
                record.data["number"]
                in self.vol_nr_dict[record.data["journal"]][record.data["volume"]]
            ):
                record.update_field(
                    key="year",
                    value=self.vol_nr_dict[record.data["journal"]][
                        record.data["volume"]
                    ][record.data["number"]],
                    source="year_vol_iss_prep",
                    note="",
                )
                record.update_masterdata_provenance(qm=self.quality_model)
        else:
            if isinstance(
                self.vol_nr_dict[record.data["journal"]][record.data["volume"]],
                (str, int),
            ):
                record.update_field(
                    key="year",
                    value=self.vol_nr_dict[record.data["journal"]][
                        record.data["volume"]
                    ],
                    source="year_vol_iss_prep",
                    note="",
                )
                record.update_masterdata_provenance(qm=self.quality_model)

    def __get_year_from_crossref(
        self, *, record: colrev.record.Record, prep_operation: colrev.ops.prep.Prep
    ) -> None:
        try:
            crossref_source = crossref_connector.CrossrefSearchSource(
                source_operation=prep_operation
            )
            retrieved_records = crossref_source.crossref_query(
                record_input=record,
                jour_vol_iss_list=True,
                timeout=prep_operation.timeout,
            )
            retries = 0
            while (
                not retrieved_records and retries < prep_operation.max_retries_on_error
            ):
                retries += 1
                retrieved_records = crossref_source.crossref_query(
                    record_input=record,
                    jour_vol_iss_list=True,
                    timeout=prep_operation.timeout,
                )
            if 0 == len(retrieved_records):
                return

            retrieved_records = [
                retrieved_record
                for retrieved_record in retrieved_records
                if retrieved_record.data.get("volume", "NA")
                == record.data.get("volume", "NA")
                and retrieved_record.data.get("journal", "NA")
                == record.data.get("journal", "NA")
                and retrieved_record.data.get("number", "NA")
                == record.data.get("number", "NA")
            ]

            years = [r.data["year"] for r in retrieved_records]
            if len(years) == 0:
                return
            most_common = max(years, key=years.count)
            if years.count(most_common) > 3:
                record.update_field(
                    key="year", value=most_common, source="CROSSREF(average)", note=""
                )
        except requests.exceptions.RequestException:
            pass

    def prepare(
        self, prep_operation: colrev.ops.prep.Prep, record: colrev.record.PrepRecord
    ) -> colrev.record.Record:
        """Prepare a record based on year-volume-issue dependency"""

        if record.data.get("year", "NA").isdigit() or record.masterdata_is_curated():
            return record

        self.__get_year_from_toc(record=record)

        if "year" in record.data:
            return record

        self.__get_year_from_vol_nr_dict(record=record)

        if "year" in record.data:
            return record

        self.__get_year_from_crossref(record=record, prep_operation=prep_operation)

        return record



#! /usr/bin/env python
"""Checker for inconsistent-with-doi-metadata."""
from __future__ import annotations

from thefuzz import fuzz

import colrev.exceptions as colrev_exceptions
import colrev.ops.built_in.search_sources.crossref as crossref_connector
import colrev.qm.quality_model

# pylint: disable=too-few-public-methods


class InconsistentWithDOIMetadataChecker:
    """The InconsistentWithDOIMetadataChecker"""

    msg = "inconsistent-with-doi-metadata"
    __fields_to_check = ["author", "title", "journal", "year", "volume", "number"]

    def __init__(self, quality_model: colrev.qm.quality_model.QualityModel) -> None:
        self.quality_model = quality_model
        self.__etiquette = crossref_connector.CrossrefSearchSource.get_etiquette(
            review_manager=quality_model.review_manager
        )

    def run(self, *, record: colrev.record.Record) -> None:
        """Run the inconsistent-with-doi-metadata checks"""

        if "doi" not in record.data:
            return
        if "doi" in record.data.get("colrev_data_provenance", {}):
            if (
                "md_curated.bib"
                in record.data["colrev_data_provenance"]["doi"]["source"]
            ):
                return

        if self.__doi_metadata_conflicts(record=record):
            record.add_masterdata_provenance_note(key="doi", note=self.msg)
        else:
            record.remove_masterdata_provenance_note(key="doi", note=self.msg)

    def __doi_metadata_conflicts(self, *, record: colrev.record.Record) -> bool:
        record_copy = record.copy_prep_rec()

        try:
            crossref_md = crossref_connector.CrossrefSearchSource.query_doi(
                doi=record_copy.data["doi"], etiquette=self.__etiquette
            )

            for key, value in crossref_md.data.items():
                if key not in self.__fields_to_check:
                    continue
                if not isinstance(value, str):
                    continue
                if key not in record.data:
                    continue
                if record.data[key] == "UNKNOWN":
                    continue
                if key not in ["author", "title", "journal"]:
                    continue
                if len(crossref_md.data[key]) < 5 or len(record.data[key]) < 5:
                    continue
                if (
                    fuzz.partial_ratio(
                        record.data[key].lower(), crossref_md.data[key].lower()
                    )
                    < 60
                ):
                    return True

        except (
            colrev_exceptions.RecordNotFoundInPrepSourceException,
            colrev_exceptions.RecordNotParsableException,
        ):
            return False
        return False


def register(quality_model: colrev.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(InconsistentWithDOIMetadataChecker(quality_model))



#! /usr/bin/env python3
"""dashboard table and graphs for synthesized records"""
from __future__ import annotations

import dash
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from dash import callback
from dash import dash_table
from dash import dcc
from dash import html
from dash import Input
from dash import Output

import colrev.record
import colrev.review_manager

dash.register_page(__name__)

review_manager = colrev.review_manager.ReviewManager()
status_operation = review_manager.get_status_operation()
records = review_manager.dataset.load_records_dict()
data = pd.DataFrame.from_records(
    [
        r
        for r in records.values()
        if r["colrev_status"] == colrev.record.RecordState.rev_synthesized
    ]
)

if data.empty:
    data = pd.DataFrame(columns=["author", "title", "year", "journal"])
else:
    data = data[["author", "title", "year", "journal"]]


def empty_figure() -> object:
    """creates an empty figure in case of invalid search"""
    figure = go.Figure(go.Scatter(x=[], y=[]))
    figure.update_layout(template=None)
    figure.update_xaxes(showgrid=False, showticklabels=False, zeroline=False)
    figure.update_yaxes(showgrid=False, showticklabels=False, zeroline=False)

    return figure


def plot_time(data_input: pd.DataFrame) -> object:
    """creates graph about papers published over time"""

    # check for data_input
    if data_input.empty:
        return empty_figure()

    # group data_input for the graph
    data2 = data_input.groupby(["year"])["year"].count().reset_index(name="count")

    # make and style the graph
    fig = px.bar(
        data2,
        x="year",
        y="count",
        template="simple_white",
        title="Profile of papers published over time",
        color="year",
    )
    fig.update_traces(marker_color="#fcb61a")

    fig.update_layout(
        title={
            "text": "<b>Profile of papers published over time</b>",
            "font": {"family": "Lato, sans-serif", "size": 30},
            "automargin": True,
            "x": 0.5,
        }
    )

    # style x axis
    fig.update_xaxes(
        title_text="Year",
        type="category",
        title_font={"family": "Lato, sans-serif", "size": 20},
        tickfont={"family": "Lato, sans-serif", "size": 20},
    )

    # style y axis
    fig.update_yaxes(
        title_text="Count",
        title_font={"family": "Lato, sans-serif", "size": 20},
        tickfont={"family": "Lato, sans-serif", "size": 20},
    )
    return fig


def plot_journals(data_input: pd.DataFrame) -> px.bar:
    """creates graph about papers published per journal"""

    # check for data_input
    if data_input.empty:
        return empty_figure()

    # group data for the graph
    data2 = data_input.groupby(["journal"])["journal"].count().reset_index(name="count")
    data2 = data2[data2["count"] != 0].sort_values(by="count", ascending=True)

    # make and style the graph
    fig = px.bar(
        data2,
        x="count",
        y="journal",
        template="simple_white",
        title="Papers Published per Journal",
        orientation="h",
    )
    fig.update_traces(marker_color="#fcb61a")

    fig.update_layout(
        title={
            "text": "<b>Papers Published per Journal</b>",
            "font": {"family": "Lato, sans-serif", "size": 30},
            "automargin": True,
            "x": 0.5,
        }
    )

    # style x axis
    fig.update_xaxes(
        title_text="Count",
        title_font={"family": "Lato, sans-serif", "size": 20},
        tickfont={"family": "Lato, sans-serif", "size": 15},
    )

    # style y axis
    fig.update_yaxes(
        title_text="Journal",
        type="category",
        title_font={"family": "Lato, sans-serif", "size": 20},
        tickfont={"family": "Lato, sans-serif", "size": 15},
    )

    return fig


# html layout for the synthesised records subpage
layout = html.Div(
    children=[
        html.Div(
            children=[
                html.Div(
                    # sorting dropdown
                    className="options",
                    children=[
                        dcc.Dropdown(
                            id="sortby",
                            options=["title", "year", "author (alphabetically)"],
                            placeholder="Sort by...",
                        )
                    ],
                ),
                html.Div(
                    # search bar
                    className="options",
                    children=[
                        dcc.Input(
                            type="text",
                            id="search",
                            value="",
                            placeholder="  Search for...",
                        )
                    ],
                ),
            ],
            className="menu",
        ),
        html.Div(
            children=[
                html.Div(
                    [
                        html.Label(
                            "Currently Synthesized Records",
                            style={
                                "font-family": "Lato, sans-serif",
                                "font-weight": "bold",
                                "fontSize": 30,
                            },
                        ),
                        # table with synthesized records
                        dash_table.DataTable(
                            data=data.to_dict("records"),
                            id="table",
                            style_cell={
                                "font-family": "Lato, sans-serif",
                                "font-size": "20px",
                                "text-align": "left",
                            },
                            style_cell_conditional=[
                                {
                                    "if": {"column_id": "year"},
                                    "width": "7%",
                                    "text-align": "center",
                                },
                                {"if": {"column_id": "title"}, "width": "53%"},
                            ],
                            style_header={
                                "font-weight": "bold",
                                "backgroundColor": "#006400",
                                "color": "white",
                            },
                            style_data={
                                "whiteSpace": "normal",
                                "height": "auto",
                                "border": "1px solid green",
                            },
                            style_as_list_view=True,
                        ),
                    ]
                ),
                # Div für Ausgabe wenn keine Ergebnisse bei Suche
                html.Div(id="table_empty", children=[]),
                # graph 1
                html.Div(
                    [dcc.Graph(figure=plot_time(data), id="time")],
                ),
                # graph 2
                html.Div(
                    [dcc.Graph(figure=plot_journals(data), id="magazines")],
                ),
            ],
            className="wrapper",
        ),
        # button to burn down chart
        html.Div(
            className="navigation-button",
            children=[
                html.A(
                    html.Button("back to Burn-Down Chart"),
                    href="http://127.0.0.1:8050/",
                )
            ],
        ),
    ]
)


@callback(
    Output("table", "data"),
    Output("table_empty", "children"),
    Output("time", "figure"),
    Output("magazines", "figure"),
    Input("search", "value"),
    Input("sortby", "value"),
)
def update_table(searchvalue, sortvalue) -> tuple[dict, str, px.bar, px.bar]:  # type: ignore
    """callback function updating table and graphs based on search and sort"""
    sorted_data = data.copy(deep=True)

    output = ""

    # sort data
    if sortvalue == "year":
        sorted_data = sorted_data.sort_values(by=["year"])  # sort by year
    elif sortvalue == "title":
        sorted_data = sorted_data.sort_values(by=["title"])  # sort by title
    elif sortvalue == "author":
        sorted_data = sorted_data.sort_values(by=["author"])  # sort by author

    data2 = sorted_data.copy(deep=True).to_dict("records")

    # search for data
    for row in sorted_data.to_dict("records"):
        found = False
        for key in row:
            if searchvalue.lower().strip() in str(row[key]).lower():
                found = True

        if found is False:
            data2.remove(row)

    # check if search results are empty
    if not data2:
        output = "No records included in the sample."

    return (
        data2,
        output,
        plot_time(pd.DataFrame.from_dict(data2)),
        plot_journals(pd.DataFrame.from_dict(data2)),
    )



#! /usr/bin/env python
"""CoLRev screen operation: Screen documents."""
from __future__ import annotations

import math
from pathlib import Path

import colrev.exceptions as colrev_exceptions
import colrev.operation
import colrev.record
import colrev.settings
import colrev.ui_cli.cli_colors as colors


class Screen(colrev.operation.Operation):
    """Screen records (based on PDFs)"""

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        notify_state_transition_operation: bool = True,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=colrev.operation.OperationsType.screen,
            notify_state_transition_operation=notify_state_transition_operation,
        )

        self.verbose = True

    def __include_all_in_screen_precondition(self, *, records: dict) -> bool:
        if not [
            r
            for r in records.values()
            if r["colrev_status"] == colrev.record.RecordState.pdf_prepared
        ]:
            if [
                r
                for r in records.values()
                if r["colrev_status"] == colrev.record.RecordState.md_processed
            ]:
                self.review_manager.logger.warning(
                    "No records to screen. Use "
                    f"{colors.ORANGE}colrev prescreen --include_all{colors.END} instead"
                )
            else:
                self.review_manager.logger.warning("No records to screen.")
            return False
        return True

    def include_all_in_screen(self, *, persist: bool) -> None:
        """Include all records in the screen"""

        if persist:
            self.review_manager.settings.screen.screen_package_endpoints = []
            self.review_manager.save_settings()

        records = self.review_manager.dataset.load_records_dict()

        if not self.__include_all_in_screen_precondition(records=records):
            return

        selected_record_ids = [
            r["ID"]
            for r in records.values()
            if colrev.record.RecordState.pdf_prepared == r["colrev_status"]
        ]

        screening_criteria = self.get_screening_criteria()

        saved_args = locals()
        saved_args["include_all"] = ""
        pad = 50
        for record_id, record in records.items():
            if record["colrev_status"] != colrev.record.RecordState.pdf_prepared:
                continue
            self.review_manager.report_logger.info(
                f" {record_id}".ljust(pad, " ") + "Included in screen (automatically)"
            )
            if len(screening_criteria) == 0:
                record.update(screening_criteria="NA")
            else:
                record.update(
                    screening_criteria=";".join([e + "=in" for e in screening_criteria])
                )
            record.update(colrev_status=colrev.record.RecordState.rev_included)

        self.review_manager.dataset.save_records_dict(records=records)
        self.review_manager.dataset.add_record_changes()
        self.__print_stats(selected_record_ids=selected_record_ids)
        self.review_manager.create_commit(
            msg="Screen (include_all)",
            manual_author=False,
        )

    def get_screening_criteria(self) -> list:
        """Get the list of screening criteria from settings"""

        return list(self.review_manager.settings.screen.criteria.keys())

    def set_screening_criteria(
        self, *, screening_criteria: dict[str, colrev.settings.ScreenCriterion]
    ) -> None:
        """Set the screening criteria in the settings"""
        self.review_manager.settings.screen.criteria = screening_criteria
        self.review_manager.save_settings()

    def get_data(self) -> dict:
        """Get the data (records to screen)"""

        # pylint: disable=duplicate-code
        records_headers = self.review_manager.dataset.load_records_dict(
            header_only=True
        )
        record_header_list = list(records_headers.values())
        nr_tasks = len(
            [
                x
                for x in record_header_list
                if colrev.record.RecordState.pdf_prepared == x["colrev_status"]
            ]
        )
        pad = 0
        if record_header_list:
            pad = min((max(len(x["ID"]) for x in record_header_list) + 2), 35)
        items = self.review_manager.dataset.read_next_record(
            conditions=[{"colrev_status": colrev.record.RecordState.pdf_prepared}]
        )
        screen_data = {"nr_tasks": nr_tasks, "PAD": pad, "items": items}
        # self.review_manager.logger.debug(
        #     self.review_manager.p_printer.pformat(screen_data)
        # )
        return screen_data

    def add_criterion(self, *, criterion_to_add: str) -> None:
        """Add a screening criterion to the records and settings"""

        assert criterion_to_add.count(",") == 2
        (
            criterion_name,
            criterion_type_str,
            criterion_explanation,
        ) = criterion_to_add.split(",")
        criterion_type = colrev.settings.ScreenCriterionType[criterion_type_str]

        records = self.review_manager.dataset.load_records_dict()

        if criterion_name not in self.review_manager.settings.screen.criteria:
            add_criterion = colrev.settings.ScreenCriterion(
                explanation=criterion_explanation,
                criterion_type=criterion_type,
                comment="",
            )
            self.review_manager.settings.screen.criteria[criterion_name] = add_criterion

            self.review_manager.save_settings()
            self.review_manager.dataset.add_setting_changes()
        else:
            print(f"Error: criterion {criterion_name} already in settings")
            return

        for record_dict in records.values():
            if record_dict["colrev_status"] in [
                colrev.record.RecordState.rev_included,
                colrev.record.RecordState.rev_synthesized,
            ]:
                record_dict["screening_criteria"] += f";{criterion_name}=TODO"
                # Note : we set the status to pdf_prepared because the screening
                # decisions have to be updated (resulting in inclusion or exclusion)
                record = colrev.record.Record(data=record_dict)
                record.set_status(target_state=colrev.record.RecordState.pdf_prepared)
            if record_dict["colrev_status"] == colrev.record.RecordState.rev_excluded:
                record_dict["screening_criteria"] += f";{criterion_name}=TODO"
                # Note : no change in colrev_status
                # because at least one of the other criteria led to exclusion decision

        self.review_manager.dataset.save_records_dict(records=records)
        self.review_manager.dataset.add_record_changes()
        self.review_manager.create_commit(
            msg=f"Add screening criterion: {criterion_name}",
        )

    def delete_criterion(self, *, criterion_to_delete: str) -> None:
        """Delete a screening criterion from the records and settings"""
        records = self.review_manager.dataset.load_records_dict()

        if criterion_to_delete in self.review_manager.settings.screen.criteria:
            del self.review_manager.settings.screen.criteria[criterion_to_delete]
            self.review_manager.save_settings()
            self.review_manager.dataset.add_setting_changes()
        else:
            print(f"Error: criterion {criterion_to_delete} not in settings")
            return

        for record_dict in records.values():
            if record_dict["colrev_status"] in [
                colrev.record.RecordState.rev_included,
                colrev.record.RecordState.rev_synthesized,
            ]:
                record_dict["screening_criteria"] = (
                    record_dict["screening_criteria"]
                    .replace(f"{criterion_to_delete}=TODO", "")
                    .replace(f"{criterion_to_delete}=in", "")
                    .replace(f"{criterion_to_delete}=out", "")
                    .replace(";;", ";")
                    .lstrip(";")
                    .rstrip(";")
                )
                # Note : colrev_status does not change
                # because the other screening criteria do not change

            if record_dict["colrev_status"] in [colrev.record.RecordState.rev_excluded]:
                record_dict["screening_criteria"] = (
                    record_dict["screening_criteria"]
                    .replace(f"{criterion_to_delete}=TODO", "")
                    .replace(f"{criterion_to_delete}=in", "")
                    .replace(f"{criterion_to_delete}=out", "")
                    .replace(";;", ";")
                    .lstrip(";")
                    .rstrip(";")
                )

                if (
                    "=out" not in record_dict["screening_criteria"]
                    and "=TODO" not in record_dict["screening_criteria"]
                ):
                    record = colrev.record.Record(data=record_dict)
                    record.set_status(
                        target_state=colrev.record.RecordState.rev_included
                    )

        self.review_manager.dataset.save_records_dict(records=records)
        self.review_manager.dataset.add_record_changes()
        self.review_manager.create_commit(
            msg=f"Removed screening criterion: {criterion_to_delete}",
        )

    def create_screen_split(self, *, create_split: int) -> list:
        """Split the screen between researchers"""

        data = self.get_data()
        nrecs = math.floor(data["nr_tasks"] / create_split)

        self.review_manager.logger.info(
            f"Creating screen splits for {create_split} researchers " f"({nrecs} each)"
        )

        screen_splits = []
        for _ in range(0, create_split):
            added: list[str] = []
            while len(added) < nrecs:
                added.append(next(data["items"])["ID"])
            addition = "colrev screen --split " + ",".join(added)
            screen_splits.append(addition)

        return screen_splits

    def setup_custom_script(self) -> None:
        """Setup a custom screen script"""

        filedata = colrev.env.utils.get_package_file_content(
            file_path=Path("template/custom_scripts/custom_screen_script.py")
        )

        if filedata:
            with open("custom_screen_script.py", "w", encoding="utf-8") as file:
                file.write(filedata.decode("utf-8"))

        self.review_manager.dataset.add_changes(path=Path("custom_screen_script.py"))

        self.review_manager.settings.screen.screen_package_endpoints.append(
            {"endpoint": "custom_screen_script"}
        )
        self.review_manager.save_settings()

    def __screen_include_all(self, *, records: dict) -> None:
        self.review_manager.logger.info("Screen: Include all records")
        for record_dict in records.values():
            if record_dict["colrev_status"] == colrev.record.RecordState.pdf_prepared:
                record = colrev.record.Record(data=record_dict)
                record.set_status(target_state=colrev.record.RecordState.rev_included)
        self.review_manager.dataset.save_records_dict(records=records)
        self.review_manager.dataset.add_record_changes()
        self.review_manager.create_commit(
            msg="Screen (include_all)",
            manual_author=False,
        )

    def __print_stats(self, *, selected_record_ids: list) -> None:
        records = self.review_manager.dataset.load_records_dict(header_only=True)
        screen_excluded = [
            r["ID"]
            for r in records.values()
            if colrev.record.RecordState.rev_excluded == r["colrev_status"]
            and r["ID"] in selected_record_ids
        ]
        screen_included = [
            r["ID"]
            for r in records.values()
            if colrev.record.RecordState.rev_included == r["colrev_status"]
            and r["ID"] in selected_record_ids
        ]

        if not screen_excluded and not screen_included:
            return

        print()
        self.review_manager.logger.info("Statistics")
        for record_dict in records.values():
            if record_dict["ID"] in screen_excluded:
                reasons = record_dict.get("screening_criteria", "NA")
                if reasons == "NA":
                    reasons = ""
                else:
                    reasons = f"({reasons})"
                self.review_manager.logger.info(
                    f" {record_dict['ID']}".ljust(40)
                    + f"pdf_prepared → rev_excluded {colors.RED}{reasons}{colors.END}"
                )
            elif record_dict["ID"] in screen_included:
                self.review_manager.logger.info(
                    f" {colors.GREEN}{record_dict['ID']}".ljust(45)
                    + f"pdf_prepared → rev_included{colors.END}"
                )

        nr_screen_excluded = len(screen_excluded)
        nr_screen_included = len(screen_included)

        self.review_manager.logger.info(
            "Excluded".ljust(29) + f"{nr_screen_excluded}".rjust(10, " ") + " records"
        )
        self.review_manager.logger.info(
            "Included".ljust(29) + f"{nr_screen_included}".rjust(10, " ") + " records"
        )

    def screen(
        self,
        *,
        record: colrev.record.Record,
        screen_inclusion: bool,
        screening_criteria: str,
        PAD: int = 40,
    ) -> None:
        """Save the screen decision"""

        PAD = 40
        if screen_inclusion:
            screening_criteria_list = self.get_screening_criteria()
            if len(screening_criteria_list) == 0:
                record.data.update(screening_criteria="NA")
            else:
                record.data.update(
                    screening_criteria=";".join(
                        [e + "=in" for e in screening_criteria_list]
                    )
                )
            record.set_status(target_state=colrev.record.RecordState.rev_included)

            self.review_manager.report_logger.info(
                f" {record.data['ID']}".ljust(PAD, " ") + "Included in screen"
            )
        else:
            record.data["screening_criteria"] = screening_criteria
            record.set_status(target_state=colrev.record.RecordState.rev_excluded)
            self.review_manager.report_logger.info(
                f" {record.data['ID']}".ljust(PAD, " ") + "Excluded in screen"
            )

        record_dict = record.get_data()
        self.review_manager.dataset.save_records_dict(
            records={record_dict["ID"]: record_dict}, partial=True
        )
        self.review_manager.dataset.add_record_changes()

    def __auto_include(self, *, records: dict) -> list:
        selected_auto_include_ids = [
            r["ID"]
            for r in records.values()
            if colrev.record.RecordState.pdf_prepared == r["colrev_status"]
            and r.get("include_flag", "0") == "1"
        ]
        if not selected_auto_include_ids:
            return selected_auto_include_ids
        self.review_manager.logger.info(
            f"{colors.GREEN}Automatically including records with include_flag{colors.END}"
        )

        for record_dict in records.values():
            if record_dict["ID"] not in selected_auto_include_ids:
                continue
            record = colrev.record.Record(data=record_dict)
            self.screen(
                record=record,
                screen_inclusion=True,
                screening_criteria="",
            )
            record.remove_field(key="include_flag")

        self.review_manager.create_commit(
            msg="Include records (include_flag)", manual_author=True
        )
        return selected_auto_include_ids

    @colrev.operation.Operation.decorate()
    def main(self, *, split_str: str = "NA") -> None:
        """Screen records for inclusion (main entrypoint)"""

        self.review_manager.logger.info("Screen")
        self.review_manager.logger.info(
            "In the screen, records are included or excluded "
            "based on the PDFs and screening criteria."
        )
        self.review_manager.logger.info(
            "See https://colrev.readthedocs.io/en/latest/manual/pdf_screen/screen.html"
        )

        # pylint: disable=duplicate-code
        split = []
        if split_str != "NA":
            split = split_str.split(",")
            if "" in split:
                split.remove("")

        records = self.review_manager.dataset.load_records_dict()

        package_manager = self.review_manager.get_package_manager()

        if not self.review_manager.settings.screen.screen_package_endpoints:
            self.__screen_include_all(records=records)
            return

        for (
            screen_package_endpoint
        ) in self.review_manager.settings.screen.screen_package_endpoints:
            endpoint_dict = package_manager.load_packages(
                package_type=colrev.env.package_manager.PackageEndpointType.screen,
                selected_packages=self.review_manager.settings.screen.screen_package_endpoints,
                operation=self,
                only_ci_supported=self.review_manager.in_ci_environment(),
            )
            if screen_package_endpoint["endpoint"] not in endpoint_dict:
                self.review_manager.logger.info(
                    f'Skip {screen_package_endpoint["endpoint"]} (not available)'
                )
                if self.review_manager.in_ci_environment():
                    raise colrev_exceptions.ServiceNotAvailableException(
                        dep="colrev sceen",
                        detailed_trace="sceen not available in ci environment",
                    )
                raise colrev_exceptions.ServiceNotAvailableException(
                    dep="colrev sceen", detailed_trace="sceen not available"
                )

            endpoint = endpoint_dict[screen_package_endpoint["endpoint"]]

            selected_auto_include_ids = self.__auto_include(records=records)

            selected_record_ids = [
                r["ID"]
                for r in records.values()
                if colrev.record.RecordState.pdf_prepared == r["colrev_status"]
                and not r.get("include_flag", "0") == "1"
            ]
            if split:
                split = [x for x in selected_record_ids if x in split]

            endpoint.run_screen(self, records, selected_record_ids)  # type: ignore

            self.__print_stats(
                selected_record_ids=selected_record_ids + selected_auto_include_ids
            )

        self.review_manager.logger.info(
            f"{colors.GREEN}Completed screen operation{colors.END}"
        )



#! /usr/bin/env python
"""CoLRev prescreen operation: Prescreen records (metadata)."""
from __future__ import annotations

import math
from pathlib import Path

import colrev.exceptions as colrev_exceptions
import colrev.operation
import colrev.ops.built_in.prescreen.conditional_prescreen
import colrev.ops.built_in.prescreen.prescreen_table
import colrev.record
import colrev.ui_cli.cli_colors as colors


class Prescreen(colrev.operation.Operation):
    """Prescreen records (based on metadata)"""

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        notify_state_transition_operation: bool = True,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=colrev.operation.OperationsType.prescreen,
            notify_state_transition_operation=notify_state_transition_operation,
        )

        self.verbose = True

    def export_table(self, *, export_table_format: str = "csv") -> None:
        """Export a table with records to prescreen"""

        endpoint = colrev.ops.built_in.prescreen.prescreen_table.TablePrescreen(
            prescreen_operation=self, settings={"endpoint": "export_table"}
        )
        records = self.review_manager.dataset.load_records_dict()
        endpoint.export_table(
            prescreen_operation=self,
            records=records,
            split=[],
            export_table_format=export_table_format,
        )

    def import_table(self, *, import_table_path: str) -> None:
        """Import a table with prescreened records"""

        endpoint = colrev.ops.built_in.prescreen.prescreen_table.TablePrescreen(
            prescreen_operation=self, settings={"endpoint": "import_table"}
        )
        records = self.review_manager.dataset.load_records_dict()
        endpoint.import_table(
            prescreen_operation=self,
            records=records,
            import_table_path=import_table_path,
        )

    def __include_all_in_prescreen_precondition(self, *, records: dict) -> bool:
        if not [
            r
            for r in records.values()
            if r["colrev_status"] == colrev.record.RecordState.md_processed
        ]:
            if [
                r
                for r in records.values()
                if r["colrev_status"] == colrev.record.RecordState.pdf_prepared
            ]:
                self.review_manager.logger.warning(
                    "No records to prescreen. Use "
                    f"{colors.ORANGE}colrev screen --include_all{colors.END} instead"
                )
            else:
                self.review_manager.logger.warning("No records to prescreen.")
            return False
        return True

    @colrev.operation.Operation.decorate()
    def include_all_in_prescreen(self, *, persist: bool) -> None:
        """Include all records in the prescreen"""

        if persist:
            self.review_manager.settings.prescreen.prescreen_package_endpoints = []
            self.review_manager.save_settings()

        records = self.review_manager.dataset.load_records_dict()
        if not self.__include_all_in_prescreen_precondition(records=records):
            return

        selected_record_ids = [
            r["ID"]
            for r in records.values()
            if colrev.record.RecordState.md_processed == r["colrev_status"]
        ]

        endpoint = (
            colrev.ops.built_in.prescreen.conditional_prescreen.ConditionalPrescreen(
                prescreen_operation=self, settings={"endpoint": "include_all"}
            )
        )
        endpoint.run_prescreen(self, records, [])
        self.__print_stats(selected_record_ids=selected_record_ids)

    def include_records(self, *, ids: str) -> None:
        """Include records in the prescreen"""

        self.__prescreen_records(ids=ids, include=True)

    def exclude_records(self, *, ids: str) -> None:
        """Exclude records in the prescreen"""

        self.__prescreen_records(ids=ids, include=False)

    def __prescreen_records(self, *, ids: str, include: bool) -> None:
        records = self.review_manager.dataset.load_records_dict()
        for record_id in ids.split(","):
            if record_id not in records:
                self.review_manager.logger.info(f" not found: {record_id}")
                continue
            record = colrev.record.Record(data=records[record_id])
            if (
                record.data["colrev_status"] != colrev.record.RecordState.md_processed
                and not self.review_manager.force_mode
            ):
                self.review_manager.logger.info(
                    f" record not md_processed / cannot prescreen: {record_id}"
                )
                continue
            if include:
                record.set_status(
                    target_state=colrev.record.RecordState.rev_prescreen_included
                )
            else:
                record.set_status(
                    target_state=colrev.record.RecordState.rev_prescreen_excluded
                )

        self.review_manager.dataset.save_records_dict(records=records)
        self.review_manager.dataset.add_record_changes()

        msg = f"Pre-screen (exclude {ids})"
        if include:
            msg = f"Pre-screen (include {ids})"

        self.review_manager.create_commit(
            msg=msg,
            manual_author=False,
        )

    def get_data(self) -> dict:
        """Get the data for prescreen"""

        # pylint: disable=duplicate-code

        records_headers = self.review_manager.dataset.load_records_dict(
            header_only=True
        )
        record_header_list = list(records_headers.values())
        nr_tasks = len(
            [
                x
                for x in record_header_list
                if colrev.record.RecordState.md_processed == x["colrev_status"]
            ]
        )
        pad = 0
        if record_header_list:
            pad = min((max(len(x["ID"]) for x in record_header_list) + 2), 40)
        items = self.review_manager.dataset.read_next_record(
            conditions=[{"colrev_status": colrev.record.RecordState.md_processed}]
        )
        prescreen_data = {"nr_tasks": nr_tasks, "PAD": pad, "items": items}

        return prescreen_data

    def create_prescreen_split(self, *, create_split: int) -> list:
        """Split the prescreen between researchers"""

        prescreen_splits = []

        data = self.get_data()
        nrecs = math.floor(data["nr_tasks"] / create_split)

        self.review_manager.report_logger.info(
            f"Creating prescreen splits for {create_split} researchers "
            f"({nrecs} each)"
        )

        for _ in range(0, create_split):
            added: list[str] = []
            while len(added) < nrecs:
                added.append(next(data["items"])["ID"])
            prescreen_splits.append("colrev prescreen --split " + ",".join(added))

        return prescreen_splits

    def setup_custom_script(self) -> None:
        """Setup a custom prescreen script"""

        filedata = colrev.env.utils.get_package_file_content(
            file_path=Path("template/custom_scripts/custom_prescreen_script.py")
        )

        if filedata:
            with open("custom_prescreen_script.py", "w", encoding="utf8") as file:
                file.write(filedata.decode("utf-8"))

        self.review_manager.dataset.add_changes(path=Path("custom_prescreen_script.py"))

        self.review_manager.settings.prescreen.prescreen_package_endpoints.append(
            {"endpoint": "custom_prescreen_script"}
        )
        self.review_manager.save_settings()

    def __prescreen_include_all(self, *, records: dict) -> None:
        # pylint: disable=duplicate-code
        self.review_manager.logger.info("Prescreen-including all records")
        for record_dict in records.values():
            if record_dict["colrev_status"] == colrev.record.RecordState.md_processed:
                record = colrev.record.Record(data=record_dict)
                record.set_status(
                    target_state=colrev.record.RecordState.rev_prescreen_included
                )
        self.review_manager.dataset.save_records_dict(records=records)
        self.review_manager.dataset.add_record_changes()
        self.review_manager.create_commit(
            msg="Pre-screen (include_all)",
            manual_author=False,
        )

    def __print_stats(self, *, selected_record_ids: list) -> None:
        records = self.review_manager.dataset.load_records_dict(header_only=True)
        prescreen_excluded = [
            r["ID"]
            for r in records.values()
            if colrev.record.RecordState.rev_prescreen_excluded == r["colrev_status"]
            and r["ID"] in selected_record_ids
        ]
        prescreen_included = [
            r["ID"]
            for r in records.values()
            if colrev.record.RecordState.rev_prescreen_included == r["colrev_status"]
            and r["ID"] in selected_record_ids
        ]

        if not prescreen_excluded and not prescreen_included:
            return

        print()
        self.review_manager.logger.info("Statistics")
        for record_dict in records.values():
            if record_dict["ID"] in prescreen_excluded:
                self.review_manager.logger.info(
                    f" {record_dict['ID']}".ljust(41)
                    + "md_processed → rev_prescreen_excluded"
                )
            elif record_dict["ID"] in prescreen_included:
                self.review_manager.logger.info(
                    f"{colors.GREEN}"
                    + f" {record_dict['ID']}".ljust(41)
                    + f"md_processed → rev_prescreen_included{colors.END}"
                )

        nr_prescreen_excluded = len(prescreen_excluded)
        nr_prescreen_included = len(prescreen_included)

        self.review_manager.logger.info(
            "Prescreen excluded".ljust(30)
            + f"{nr_prescreen_excluded}".rjust(10, " ")
            + " records"
        )
        self.review_manager.logger.info(
            "Prescreen included".ljust(30)
            + f"{nr_prescreen_included}".rjust(10, " ")
            + " records"
        )

    def prescreen(
        self,
        *,
        record: colrev.record.Record,
        prescreen_inclusion: bool,
        PAD: int = 40,
    ) -> None:
        """Save the prescreen decision"""
        if prescreen_inclusion:
            self.review_manager.report_logger.info(
                f" {record.data['ID']}".ljust(PAD, " ") + "Included in prescreen"
            )
            record.set_status(
                target_state=colrev.record.RecordState.rev_prescreen_included
            )
            self.review_manager.dataset.save_records_dict(
                records={record.data["ID"]: record.get_data()}, partial=True
            )

        else:
            self.review_manager.report_logger.info(
                f" {record.data['ID']}".ljust(PAD, " ") + "Excluded in prescreen"
            )
            record.set_status(
                target_state=colrev.record.RecordState.rev_prescreen_excluded
            )
            self.review_manager.dataset.save_records_dict(
                records={record.data["ID"]: record.get_data()}, partial=True
            )

        self.review_manager.dataset.add_record_changes()

    def __auto_include(self, *, records: dict) -> list:
        selected_auto_include_ids = [
            r["ID"]
            for r in records.values()
            if colrev.record.RecordState.md_processed == r["colrev_status"]
            and r.get("include_flag", "0") == "1"
        ]
        if not selected_auto_include_ids:
            return selected_auto_include_ids
        self.review_manager.logger.info(
            f"{colors.GREEN}Automatically including records with include_flag{colors.END}"
        )
        for record_dict in records.values():
            if record_dict["ID"] not in selected_auto_include_ids:
                continue
            self.prescreen(
                record=colrev.record.Record(data=record_dict),
                prescreen_inclusion=True,
            )
        return selected_auto_include_ids

    @colrev.operation.Operation.decorate()
    def main(self, *, split_str: str = "NA") -> None:
        """Prescreen records (main entrypoint)"""

        # pylint: disable=duplicate-code
        split = []
        if split_str != "NA":
            split = split_str.split(",")
            if "" in split:
                split.remove("")

        records = self.review_manager.dataset.load_records_dict()

        package_manager = self.review_manager.get_package_manager()

        prescreen_package_endpoints = (
            self.review_manager.settings.prescreen.prescreen_package_endpoints
        )

        if not prescreen_package_endpoints:
            self.__prescreen_include_all(records=records)
            return

        for prescreen_package_endpoint in prescreen_package_endpoints:
            self.review_manager.logger.debug(
                f"Run {prescreen_package_endpoint['endpoint']}"
            )
            endpoint_dict = package_manager.load_packages(
                package_type=colrev.env.package_manager.PackageEndpointType.prescreen,
                selected_packages=prescreen_package_endpoints,
                operation=self,
                only_ci_supported=self.review_manager.in_ci_environment(),
            )
            if prescreen_package_endpoint["endpoint"] not in endpoint_dict:
                self.review_manager.logger.info(
                    f'Skip {prescreen_package_endpoint["endpoint"]} (not available)'
                )
                if self.review_manager.in_ci_environment():
                    raise colrev_exceptions.ServiceNotAvailableException(
                        dep="colrev presceen",
                        detailed_trace="presceen not available in ci environment",
                    )
                raise colrev_exceptions.ServiceNotAvailableException(
                    dep="colrev presceen", detailed_trace="presceen not available"
                )

            endpoint = endpoint_dict[prescreen_package_endpoint["endpoint"]]

            selected_record_ids = [
                r["ID"]
                for r in records.values()
                if colrev.record.RecordState.md_processed == r["colrev_status"]
                and not r.get("include_flag", "0") == "1"
            ]
            endpoint.run_prescreen(self, records, split)  # type: ignore

            selected_auto_include_ids = self.__auto_include(records=records)

            self.__print_stats(
                selected_record_ids=selected_record_ids + selected_auto_include_ids
            )

        self.review_manager.logger.info(
            "%sCompleted prescreen operation%s", colors.GREEN, colors.END
        )



#! /usr/bin/env python
"""OCR as a PDF preparation operation"""
from __future__ import annotations

import os
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import docker
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.language_service
import colrev.env.package_manager
import colrev.env.utils
import colrev.record

if TYPE_CHECKING:
    import colrev.ops.pdf_prep

# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.env.package_manager.PDFPrepPackageEndpointInterface)
@dataclass
class PDFCheckOCR(JsonSchemaMixin):
    """Prepare PDFs by checking and applying OCR (if necessary) based on OCRmyPDF"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = False

    def __init__(
        self,
        *,
        pdf_prep_operation: colrev.ops.pdf_prep.PDFPrep,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

        if not pdf_prep_operation.review_manager.in_ci_environment():
            self.ocrmypdf_image = "jbarlow83/ocrmypdf:latest"
            pdf_prep_operation.review_manager.environment_manager.build_docker_image(
                imagename=self.ocrmypdf_image
            )
        self.language_service = colrev.env.language_service.LanguageService()

    def __text_is_english(self, *, text: str) -> bool:
        return "eng" == self.language_service.compute_language(text=text)

    def __apply_ocr(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        record: colrev.record.Record,
        pad: int,  # pylint: disable=unused-argument
    ) -> colrev.record.Record:
        pdf_path = review_manager.path / Path(record.data["file"])
        non_ocred_filename = Path(str(pdf_path).replace(".pdf", "_no_ocr.pdf"))
        pdf_path.rename(non_ocred_filename)
        orig_path = (
            pdf_path.parents[0] if pdf_path.is_file() else review_manager.pdf_dir
        )

        # options = ""
        # if rotate:
        #     options = options + '--rotate-pages '
        # if deskew:
        #     options = options + '--deskew '
        docker_home_path = Path("/home/docker")

        args = (
            f"--force-ocr --jobs 4 -l eng {str(docker_home_path / non_ocred_filename.name)} "
            f"{str(docker_home_path / pdf_path.name)}"
        )

        client = docker.from_env(timeout=120)
        client.containers.run(
            image=self.ocrmypdf_image,
            command=args,
            auto_remove=True,
            user=f"{os.geteuid()}:{os.getegid()}",
            volumes=[f"{orig_path}:/home/docker"],
        )

        record.add_data_provenance_note(key="file", note="pdf_processed with OCRMYPDF")
        record.set_text_from_pdf(project_path=review_manager.path)
        return record

    def prep_pdf(
        self,
        pdf_prep_operation: colrev.ops.pdf_prep.PDFPrep,
        record: colrev.record.Record,
        pad: int,
    ) -> dict:
        """Prepare the PDF by checking/applying OCR"""

        if colrev.record.RecordState.pdf_imported != record.data.get(
            "colrev_status", "NA"
        ):
            return record.data

        if not record.data["file"].endswith(".pdf"):
            return record.data

        # We may allow for other languages in this and the following if statement
        if not self.__text_is_english(text=record.data["text_from_pdf"]):
            pdf_prep_operation.review_manager.report_logger.info(
                f'apply_ocr({record.data["ID"]})'
            )
            record = self.__apply_ocr(
                review_manager=pdf_prep_operation.review_manager,
                record=record,
                pad=pad,
            )

        if not self.__text_is_english(text=record.data["text_from_pdf"]):
            msg = (
                f'{record.data["ID"]}'.ljust(pad, " ")
                + "Validation error (Language not English or OCR problems)"
            )
            pdf_prep_operation.review_manager.report_logger.error(msg)
            record.add_data_provenance_note(key="file", note="pdf_language_not_english")
            record.data.update(
                colrev_status=colrev.record.RecordState.pdf_needs_manual_preparation
            )
        return record.data



#! /usr/bin/env python
"""Conditional prescreen"""
from __future__ import annotations

import typing
from dataclasses import dataclass

import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.record

if typing.TYPE_CHECKING:
    import colrev.ops.prescreen.Prescreen

# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.PrescreenPackageEndpointInterface
)
@dataclass
class ConditionalPrescreen(JsonSchemaMixin):

    """Conditional prescreen (currently: include all)"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = True

    def __init__(
        self,
        *,
        prescreen_operation: colrev.ops.prescreen.Prescreen,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

    def run_prescreen(
        self,
        prescreen_operation: colrev.ops.prescreen.Prescreen,
        records: dict,
        split: list,  # pylint: disable=unused-argument
    ) -> dict:
        """Prescreen records based on predefined conditions (rules)"""

        pad = 50
        for record in records.values():
            if record["colrev_status"] != colrev.record.RecordState.md_processed:
                continue
            prescreen_operation.review_manager.report_logger.info(
                f' {record["ID"]}'.ljust(pad, " ")
                + "Included in prescreen (automatically)"
            )
            record.update(
                colrev_status=colrev.record.RecordState.rev_prescreen_included
            )

        prescreen_operation.review_manager.dataset.save_records_dict(records=records)
        prescreen_operation.review_manager.dataset.add_record_changes()
        prescreen_operation.review_manager.create_commit(
            msg="Pre-screen (include_all)",
            manual_author=False,
        )
        return records



#! /usr/bin/env python
"""Structured data extraction as part of the data operations"""
from __future__ import annotations

import csv
import typing
from dataclasses import asdict
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import pandas as pd
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin
from git.exc import GitCommandError

import colrev.env.package_manager
import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.record
import colrev.ui_cli.cli_colors as colors

if TYPE_CHECKING:
    import colrev.ops.data


# an option: https://pypi.org/project/csv-schema/
@dataclass
class Field(JsonSchemaMixin):
    """Field definition"""

    name: str
    explanation: str
    data_type: str


@zope.interface.implementer(colrev.env.package_manager.DataPackageEndpointInterface)
@dataclass
class StructuredData(JsonSchemaMixin):
    """Summarize the literature in a structured data extraction (a table)"""

    settings: StructuredDataSettings
    ci_supported: bool = False

    @dataclass
    class StructuredDataSettings(
        colrev.env.package_manager.DefaultSettings, JsonSchemaMixin
    ):
        """Settings for StructuredData"""

        endpoint: str
        version: str
        fields: typing.List[Field]
        data_path_relative: Path = Path("data.csv")

        _details = {
            "fields": {"tooltip": "Fields for the structured data extraction"},
        }

    settings_class = StructuredDataSettings

    __FULL_DATA_FIELD_EXPLANATION = """Explanation: Data fields are used in the coding tables.
Example 1:
    - name           : summary
    - explanation    : Brief summary of the principal findings
    - data_type      : str

Example 2:
    - name           : sample_size
    - explanation    : Sample size of the study
    - data_type      : int
    """

    def __init__(
        self,
        *,
        data_operation: colrev.ops.data.Data,
        settings: dict,
    ) -> None:
        if "version" not in settings:
            settings["version"] = "0.1"

        if "fields" not in settings:
            settings["fields"] = []
        if "data_path_relative" not in settings:
            settings["data_path_relative"] = Path("data.csv")

        self.settings = self.settings_class.load_settings(data=settings)
        self.data_path = (
            data_operation.review_manager.data_dir / self.settings.data_path_relative
        )
        self.review_manager = data_operation.review_manager

    # pylint: disable=unused-argument
    @classmethod
    def add_endpoint(cls, operation: colrev.ops.data.Data, params: str) -> None:
        """Add as an endpoint"""

        add_source = {
            "endpoint": "colrev.structured",
            "version": "0.1",
            "fields": [],
            "data_path_relative": "data/data.csv",
        }
        operation.review_manager.settings.data.data_package_endpoints.append(add_source)

    def validate_structured_data(self) -> None:
        """Validate the extracted data"""

        if not self.data_path.is_file():
            return

        data_df = pd.read_csv(self.data_path, dtype=str)

        # Check for duplicate IDs
        if not data_df["ID"].is_unique:
            raise colrev_exceptions.DataException(
                msg=f"duplicates in {self.settings.data_path_relative}: "
                + ",".join(data_df[data_df.duplicated(["ID"])].ID.tolist())
            )

        # Check consistency: data -> inclusion_2
        data_ids = data_df["ID"].tolist()
        records = self.review_manager.dataset.load_records_dict()
        for data_id in data_ids:
            if data_id not in records:
                raise colrev_exceptions.DataException(
                    msg=f"{data_id} in {self.settings.data_path_relative} not in records"
                )
            if records[data_id]["colrev_status"] not in [
                colrev.record.RecordState.rev_included,
                colrev.record.RecordState.rev_synthesized,
            ]:
                raise colrev_exceptions.DataException(
                    msg=f"{data_id} in {self.settings.data_path_relative} "
                    + "not in [rev_included, rev_synthesized]"
                )

        # Note : missing IDs are added through update_data

        return

    def __set_fields(self) -> None:
        self.review_manager.logger.info("Add fields for data extraction")
        try:
            _ = self.review_manager.dataset.get_repo()
        except GitCommandError:
            return

        print(self.__FULL_DATA_FIELD_EXPLANATION)
        while "y" == input("Add a data field [y,n]?"):
            short_name = input("Provide a short name: ")
            explanation = input("Provide a short explanation: ")
            data_type = input("Provide the data type (str, int, double):")

            self.settings.fields.append(
                Field(
                    **{
                        "name": short_name,
                        "explanation": explanation,
                        "data_type": data_type,
                    }
                )
            )

            print()

        # Note : the following should be easier...
        for i, candidate in enumerate(
            self.review_manager.settings.data.data_package_endpoints
        ):
            if candidate.get("data_path_relative", "NA") == str(
                self.settings.data_path_relative
            ):
                self.review_manager.settings.data.data_package_endpoints[i] = asdict(
                    self.settings, dict_factory=colrev.env.utils.custom_asdict_factory
                )
        self.review_manager.save_settings()

    def update_data(
        self,
        data_operation: colrev.ops.data.Data,
        records: dict,
        synthesized_record_status_matrix: dict,
        silent_mode: bool,
    ) -> None:
        """Update the data/structured data extraction"""

        def update_structured_data(
            *,
            review_manager: colrev.review_manager.ReviewManager,
            synthesized_record_status_matrix: dict,
        ) -> typing.Dict:
            if not self.data_path.is_file():
                self.__set_fields()

                field_names = [f["name"] for f in self.settings.fields]
                data_df = pd.DataFrame([], columns=["ID"] + field_names)
                data_df.sort_values(by=["ID"], inplace=True)

                data_df.to_csv(self.data_path, index=False, quoting=csv.QUOTE_ALL)

            nr_records_added = 0

            if not silent_mode:
                self.review_manager.report_logger.info("Update structured data")
                self.review_manager.logger.info(
                    f"Update structured data ({self.settings.data_path_relative})"
                )

            data_df = pd.read_csv(self.data_path, dtype=str)

            for record_id in list(synthesized_record_status_matrix.keys()):
                # skip when already available
                if 0 < len(data_df[data_df["ID"].str.startswith(record_id)]):
                    continue

                add_record = pd.DataFrame({"ID": [record_id]})
                add_record = add_record.reindex(
                    columns=data_df.columns, fill_value="TODO"
                )
                data_df = pd.concat([data_df, add_record], axis=0, ignore_index=True)
                review_manager.logger.info(
                    f" {colors.GREEN}{record_id}".ljust(45)
                    + f"add to structured_data{colors.END}"
                )
                nr_records_added = nr_records_added + 1

            data_df.sort_values(by=["ID"], inplace=True)

            data_df.to_csv(self.data_path, index=False, quoting=csv.QUOTE_ALL)

            if not (0 == nr_records_added and silent_mode):
                review_manager.logger.info(
                    f"Added to {self.settings.data_path_relative}".ljust(24)
                    + f"{nr_records_added}".rjust(15, " ")
                    + " records"
                )

                review_manager.logger.info(
                    f"Added to {self.settings.data_path_relative}".ljust(24)
                    + f"{nr_records_added}".rjust(15, " ")
                    + " records"
                )
            return records

        self.validate_structured_data()
        records = update_structured_data(
            review_manager=data_operation.review_manager,
            synthesized_record_status_matrix=synthesized_record_status_matrix,
        )

        data_operation.review_manager.dataset.add_changes(
            path=self.settings.data_path_relative
        )

    def update_record_status_matrix(
        self,
        data_operation: colrev.ops.data.Data,  # pylint: disable=unused-argument
        synthesized_record_status_matrix: dict,
        endpoint_identifier: str,
    ) -> None:
        """Update the record_status_matrix"""

        def get_data_extracted(
            data_path: Path, records_for_data_extraction: list
        ) -> list:
            data_extracted = []
            data_df = pd.read_csv(data_path)

            for record in records_for_data_extraction:
                drec = data_df.loc[data_df["ID"] == record]
                if 1 == drec.shape[0]:
                    if "TODO" not in drec.iloc[0].tolist():
                        data_extracted.append(drec.loc[drec.index[0], "ID"])

            data_extracted = [
                x for x in data_extracted if x in records_for_data_extraction
            ]
            return data_extracted

        def get_structured_data_extracted(
            *, synthesized_record_status_matrix: typing.Dict, data_path: Path
        ) -> list:
            if not data_path.is_file():
                return []

            data_extracted = get_data_extracted(
                data_path, list(synthesized_record_status_matrix.keys())
            )
            data_extracted = [
                x
                for x in data_extracted
                if x in list(synthesized_record_status_matrix.keys())
            ]
            return data_extracted

        structured_data_extracted = get_structured_data_extracted(
            synthesized_record_status_matrix=synthesized_record_status_matrix,
            data_path=self.data_path,
        )

        for syn_id in structured_data_extracted:
            if syn_id in synthesized_record_status_matrix:
                synthesized_record_status_matrix[syn_id][endpoint_identifier] = True
            else:
                print(f"Error: {syn_id} not int " f"{synthesized_record_status_matrix}")

    def get_advice(
        self,
        review_manager: colrev.review_manager.ReviewManager,  # pylint: disable=unused-argument
    ) -> dict:
        """Get advice on the next steps (for display in the colrev status)"""

        data_endpoint = "Data operation [structured data endpoint]: "

        if self.settings.data_path_relative.is_file():
            advice = {
                "msg": f"{data_endpoint}"
                + f"\n    - Complete the data extraction ({self.settings.data_path_relative})",
                "detailed_msg": "TODO",
            }
        else:
            advice = {}
        return advice



#! /usr/bin/env python
"""Creation of a github-page for the review as part of the data operations"""
from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import git
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.env.utils
import colrev.record
import colrev.ui_cli.cli_colors as colors

if TYPE_CHECKING:
    import colrev.ops.data


@zope.interface.implementer(colrev.env.package_manager.DataPackageEndpointInterface)
@dataclass
class GithubPages(JsonSchemaMixin):
    """Export the literature review into a Github Page"""

    settings: GHPagesSettings

    ci_supported: bool = False

    @dataclass
    class GHPagesSettings(colrev.env.package_manager.DefaultSettings, JsonSchemaMixin):
        """Settings for GithubPages"""

        endpoint: str
        version: str
        auto_push: bool

        _details = {
            "auto_push": {
                "tooltip": "Indicates whether the Github Pages branch "
                "should be pushed automatically"
            },
        }

    GH_PAGES_BRANCH_NAME = "gh-pages"

    settings_class = GHPagesSettings

    def __init__(
        self,
        *,
        data_operation: colrev.ops.data.Data,
        settings: dict,
    ) -> None:
        # Set default values (if necessary)
        if "version" not in settings:
            settings["version"] = "0.1.0"
        if "auto_push" not in settings:
            settings["auto_push"] = True

        self.settings = self.settings_class.load_settings(data=settings)
        self.review_manager = data_operation.review_manager

    # pylint: disable=unused-argument
    @classmethod
    def add_endpoint(cls, operation: colrev.ops.data.Data, params: str) -> None:
        """Add as an endpoint"""

        add_source = {
            "endpoint": "colrev.github_pages",
            "version": "0.1",
            "auto_push": True,
        }

        operation.review_manager.settings.data.data_package_endpoints.append(add_source)

    def __setup_github_pages_branch(
        self, *, data_operation: colrev.ops.data.Data, git_repo: git.Repo
    ) -> None:
        # if branch does not exist: create and add index.html
        data_operation.review_manager.logger.info("Setup gh-pages branch")
        git_repo.git.checkout("--orphan", self.GH_PAGES_BRANCH_NAME)
        git_repo.git.rm("-rf", Path("."))

        colrev.env.utils.retrieve_package_file(
            template_file=Path("template/github_pages/README.md"),
            target=Path("README.md"),
        )
        project_title = data_operation.review_manager.settings.project.title
        colrev.env.utils.inplace_change(
            filename=Path("README.md"),
            old_string="{{project_title}}",
            new_string=project_title.rstrip(" ").capitalize(),
        )
        data_operation.review_manager.dataset.add_changes(path=Path("README.md"))

        colrev.env.utils.retrieve_package_file(
            template_file=Path("template/github_pages/index.html"),
            target=Path("index.html"),
        )
        data_operation.review_manager.dataset.add_changes(path=Path("index.html"))

        colrev.env.utils.retrieve_package_file(
            template_file=Path("template/github_pages/_config.yml"),
            target=Path("_config.yml"),
        )
        colrev.env.utils.inplace_change(
            filename=Path("_config.yml"),
            old_string="{{project_title}}",
            new_string=project_title,
        )
        data_operation.review_manager.dataset.add_changes(path=Path("_config.yml"))

        colrev.env.utils.retrieve_package_file(
            template_file=Path("template/github_pages/about.md"),
            target=Path("about.md"),
        )
        data_operation.review_manager.dataset.add_changes(path=Path("about.md"))

        data_operation.review_manager.create_commit(
            msg="Setup gh-pages branch", skip_status_yaml=True
        )

    def __update_data(
        self,
        *,
        data_operation: colrev.ops.data.Data,
        git_repo: git.Repo,
        silent_mode: bool,
    ) -> None:
        if not silent_mode:
            data_operation.review_manager.logger.info("Update data on github pages")

        records = data_operation.review_manager.dataset.load_records_dict()

        included_records = {
            r["ID"]: r
            for r in records.values()
            if r["colrev_status"]
            in [
                colrev.record.RecordState.rev_synthesized,
                colrev.record.RecordState.rev_included,
            ]
        }

        git_repo.git.checkout(self.GH_PAGES_BRANCH_NAME)
        if not Path("pre-commit-config.yaml").is_file():
            colrev.env.utils.retrieve_package_file(
                template_file=Path("template/github_pages/pre-commit-config.yaml"),
                target=Path(".pre-commit-config.yaml"),
            )
            data_operation.review_manager.dataset.add_changes(
                path=Path(".pre-commit-config.yaml")
            )

        data_file = Path("data.bib")
        data_operation.review_manager.dataset.save_records_dict_to_file(
            records=included_records, save_path=data_file
        )
        data_operation.review_manager.dataset.add_changes(path=data_file)

        data_operation.review_manager.create_commit(
            msg="Update sample", skip_status_yaml=True
        )

    def __push_branch(
        self,
        *,
        data_operation: colrev.ops.data.Data,
        git_repo: git.Repo,
        silent_mode: bool,
    ) -> None:
        if not silent_mode:
            data_operation.review_manager.logger.info("Push to github pages")
        if "origin" in git_repo.remotes:
            if "origin/gh-pages" in [r.name for r in git_repo.remotes.origin.refs]:
                try:
                    git_repo.remotes.origin.push(
                        refspec=f"{self.GH_PAGES_BRANCH_NAME}:{self.GH_PAGES_BRANCH_NAME}"
                    )
                except git.exc.GitCommandError:  # pylint: disable=no-member
                    data_operation.review_manager.logger.error(
                        "Could not push branch gh-pages. Please resolve manually, i.e., run "
                        f"{colors.ORANGE}git switch gh-pages && "
                        f"git pull --rebase && git push{colors.END}"
                    )
            else:
                git_repo.git.push(
                    "--set-upstream",
                    "origin",
                    self.GH_PAGES_BRANCH_NAME,
                    "--no-verify",
                )

            username, project = (
                git_repo.remotes.origin.url.replace("https://github.com/", "")
                .replace(".git", "")
                .split("/")
            )
            if not silent_mode:
                data_operation.review_manager.logger.info(
                    f"Data available at: https://{username}.github.io/{project}/"
                )
        else:
            if not silent_mode:
                data_operation.review_manager.logger.info("No remotes specified")

    def __check_gh_pages_setup(self, *, git_repo: git.Repo) -> None:
        username, project = (
            git_repo.remotes.origin.url.replace("https://github.com/", "")
            .replace(".git", "")
            .split("/")
        )
        gh_page_link = f"https://{username}.github.io/{project}/"
        if Path("readme.md").is_file():
            if gh_page_link in (self.review_manager.path / Path("readme.md")).read_text(
                encoding="utf-8"
            ):
                return
        print(
            f"{colors.ORANGE}The Github page is not yet linked in the readme.md file.\n"
            "To make it easier to access the page, add the following to the readme.md file:\n"
            f"\n    [Github page]({gh_page_link}){colors.END}\n"
        )

    def update_data(
        self,
        data_operation: colrev.ops.data.Data,
        records: dict,  # pylint: disable=unused-argument
        synthesized_record_status_matrix: dict,  # pylint: disable=unused-argument
        silent_mode: bool,
    ) -> None:
        """Update the data/github pages"""

        # pylint: disable=too-many-branches

        if data_operation.review_manager.in_ci_environment():
            data_operation.review_manager.logger.error(
                "Running in CI environment. Skipping github-pages generation."
            )
            return

        if data_operation.review_manager.dataset.has_changes():
            data_operation.review_manager.logger.error(
                "Cannot update github pages because there are uncommited changes."
            )
            return

        git_repo = data_operation.review_manager.dataset.get_repo()
        active_branch = git_repo.active_branch

        # check if there is an "origin" remote

        if "origin" in git_repo.remotes:
            # check if remote.origin has a gh-pages branch
            git_repo.remotes.origin.fetch()
            if f"origin/{self.GH_PAGES_BRANCH_NAME}" in [
                r.name for r in git_repo.remotes.origin.refs
            ]:
                try:
                    if "origin/gh-pages" not in [
                        r.name for r in git_repo.remotes.origin.refs
                    ]:
                        git_repo.git.push(
                            "--set-upstream",
                            "origin",
                            self.GH_PAGES_BRANCH_NAME,
                            "--no-verify",
                        )
                    else:
                        git_repo.remotes.origin.pull(ff_only=True)

                    # update
                    git_repo.git.checkout(active_branch)
                    self.__update_data(
                        data_operation=data_operation,
                        git_repo=git_repo,
                        silent_mode=silent_mode,
                    )

                    # Push gh-pages branch to remote origin
                    if self.settings.auto_push:
                        self.__push_branch(
                            data_operation=data_operation,
                            git_repo=git_repo,
                            silent_mode=silent_mode,
                        )
                except git.exc.GitCommandError as exc:  # pylint: disable=no-member
                    print("Error in gh-pages:")
                    print(exc)
                    return
            else:
                # create branch
                if self.GH_PAGES_BRANCH_NAME not in [h.name for h in git_repo.heads]:
                    self.__setup_github_pages_branch(
                        data_operation=data_operation, git_repo=git_repo
                    )

                # update
                git_repo.git.checkout(active_branch)
                self.__update_data(
                    data_operation=data_operation,
                    git_repo=git_repo,
                    silent_mode=silent_mode,
                )

                # Push gh-pages branch to remote origin
                if self.settings.auto_push:
                    self.__push_branch(
                        data_operation=data_operation,
                        git_repo=git_repo,
                        silent_mode=silent_mode,
                    )
            git_repo.git.checkout(active_branch)
            self.__check_gh_pages_setup(git_repo=git_repo)
        else:
            data_operation.review_manager.logger.warning(
                "Cannot push github pages because there is no remote origin. "
                "gh-pages branch will only be created locally."
            )
            # create branch
            if self.GH_PAGES_BRANCH_NAME not in [h.name for h in git_repo.heads]:
                self.__setup_github_pages_branch(
                    data_operation=data_operation, git_repo=git_repo
                )

            # update
            git_repo.git.checkout(active_branch)
            self.__update_data(
                data_operation=data_operation,
                git_repo=git_repo,
                silent_mode=silent_mode,
            )

        git_repo.git.checkout(active_branch)

    def update_record_status_matrix(
        self,
        data_operation: colrev.ops.data.Data,  # pylint: disable=unused-argument
        synthesized_record_status_matrix: dict,
        endpoint_identifier: str,
    ) -> None:
        """Update the record_status_matrix"""

        # Note : automatically set all to True / synthesized
        for syn_id in synthesized_record_status_matrix:
            synthesized_record_status_matrix[syn_id][endpoint_identifier] = True

    def get_advice(
        self,
        review_manager: colrev.review_manager.ReviewManager,
    ) -> dict:
        """Get advice on the next steps (for display in the colrev status)"""

        data_endpoint = "Data operation [github pages data endpoint]: "

        advice = {"msg": f"{data_endpoint}", "detailed_msg": "TODO"}
        if review_manager.dataset.get_remote_url() == "NA":
            advice["msg"] += (
                "\n    - To make the repository available on Github pages, "
                + "push it to a Github repository\nhttps://github.com/new"
            )
        else:
            advice[
                "msg"
            ] += "\n    - The page is updated automatically (gh-pages branch)"

        return advice



#! /usr/bin/env python
"""Checker for inconsistent-content."""
from __future__ import annotations

import colrev.qm.quality_model

# pylint: disable=too-few-public-methods


class InconsistentContentChecker:
    """The InconsistentContentChecker"""

    msg = "inconsistent-content"

    def __init__(self, quality_model: colrev.qm.quality_model.QualityModel) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.Record) -> None:
        """Run the inconsistent-content checks"""

        for key in ["journal", "booktitle", "author"]:
            if key not in record.data:
                continue

            if self.__inconsistent_content(record=record, key=key):
                record.add_masterdata_provenance_note(key=key, note=self.msg)
            else:
                record.remove_masterdata_provenance_note(key=key, note=self.msg)

    def __inconsistent_content(self, *, record: colrev.record.Record, key: str) -> bool:
        if key == "journal":
            if "journal" in record.data and any(
                x in record.data["journal"].lower() for x in ["conference", "workshop"]
            ):
                return True
        if key == "booktitle":
            if "booktitle" in record.data and any(
                x in record.data["booktitle"].lower() for x in ["journal"]
            ):
                return True

        return False


def register(quality_model: colrev.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(InconsistentContentChecker(quality_model))



#! /usr/bin/env python
"""SearchSource: ABI/INFORM (ProQuest)"""
from __future__ import annotations

import re
from dataclasses import dataclass
from pathlib import Path

import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.load_utils_bib
import colrev.ops.search
import colrev.record

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class ABIInformProQuestSearchSource(JsonSchemaMixin):
    """SearchSource for ABI/INFORM (ProQuest)"""

    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "{{ID}}"
    search_type = colrev.settings.SearchType.DB
    api_search_supported = False
    ci_supported: bool = False
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.supported
    short_name = "ABI/INFORM (ProQuest)"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/colrev/"
        + "ops/built_in/search_sources/abi_inform_proquest.md"
    )

    def __init__(
        self, *, source_operation: colrev.operation.Operation, settings: dict
    ) -> None:
        self.search_source = from_dict(data_class=self.settings_class, data=settings)
        self.quality_model = source_operation.review_manager.get_qm()
        self.review_manager = source_operation.review_manager

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for ABI/INFORM (ProQuest)"""

        result = {"confidence": 0.0}

        if "www.proquest.com" in data:  # nosec
            if data.count("www.proquest.com") >= data.count("\n@"):
                result["confidence"] = 1.0

        return result

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.search.Search, params: str) -> None:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""
        raise NotImplementedError

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of ABI/INFORM"""

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Not implemented"""
        return record

    def __remove_duplicates(self, *, records: dict) -> None:
        to_delete = []
        for record in records.values():
            if re.search(r"-\d{1,2}$", record["ID"]):
                original_record_id = re.sub(r"-\d{1,2}$", "", record["ID"])
                if original_record_id not in records:
                    continue
                original_record = records[original_record_id]

                # Note: between duplicate records,
                # there are variations in spelling and completeness
                if (
                    colrev.record.Record.get_record_similarity(
                        record_a=colrev.record.Record(data=record),
                        record_b=colrev.record.Record(data=original_record),
                    )
                    < 0.9
                ):
                    continue

                if original_record_id not in records:
                    continue
                to_delete.append(record["ID"])
        if to_delete:
            for rid in to_delete:
                self.review_manager.logger.info(f" remove duplicate {rid}")
                del records[rid]

            self.review_manager.dataset.save_records_dict_to_file(
                records=records, save_path=self.search_source.filename
            )

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.ops.load_utils_bib.load_bib_file(
                load_operation=load_operation, source=self.search_source
            )
            self.__remove_duplicates(records=records)
            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for ABI/INFORM (ProQuest)"""

        if record.data.get("journal", "").lower().endswith("conference proceedings."):
            record.change_entrytype(
                new_entrytype="inproceedings", qm=self.quality_model
            )

        if "language" in record.data:
            if record.data["language"] in ["ENG", "English"]:
                record.update_field(
                    key="language",
                    value="eng",
                    source="prep_abi_inform_proquest_source",
                )

        return record



#! /usr/bin/env python
"""Theoretical review"""
from dataclasses import dataclass

import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.search
import colrev.record

# pylint: disable=unused-argument
# pylint: disable=duplicate-code
# pylint: disable=too-few-public-methods


@zope.interface.implementer(
    colrev.env.package_manager.ReviewTypePackageEndpointInterface
)
@dataclass
class TheoreticalReview(JsonSchemaMixin):
    """Theoretical review"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = True

    def __init__(
        self, *, operation: colrev.operation.CheckOperation, settings: dict
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

    def __str__(self) -> str:
        return "theoretical review"

    def initialize(
        self, settings: colrev.settings.Settings
    ) -> colrev.settings.Settings:
        """Initialize a theoretical review"""

        settings.data.data_package_endpoints = [
            {
                "endpoint": "colrev.obsidian",
                "version": "0.1",
                "config": {},
            },
            {
                "endpoint": "colrev.paper_md",
                "version": "1.0",
                "word_template": "APA-7.docx",
            },
        ]
        return settings
